import logging
import os
import re
from textwrap import dedent

import pytest

from ufo2ft.constants import ANCHOR_LIB_GPOS_CONTEXT_KEY, OBJECT_LIBS_KEY
from ufo2ft.featureCompiler import FeatureCompiler, parseLayoutFeatures
from ufo2ft.featureWriters import ast
from ufo2ft.featureWriters.markFeatureWriter import (
    MarkFeatureWriter,
    NamedAnchor,
    parseAnchorName,
)

from . import FeatureWriterTest


@pytest.fixture
def testufo(FontClass):
    ufo = FontClass()
    ufo.newGlyph("a").appendAnchor({"name": "top", "x": 100, "y": 200})
    liga = ufo.newGlyph("f_i")
    liga.appendAnchor({"name": "top_1", "x": 100, "y": 500})
    liga.appendAnchor({"name": "top_2", "x": 600, "y": 500})
    ufo.newGlyph("acutecomb").appendAnchor({"name": "_top", "x": 100, "y": 200})
    accent = ufo.newGlyph("tildecomb")
    accent.appendAnchor({"name": "_top", "x": 100, "y": 200})
    accent.appendAnchor({"name": "top", "x": 100, "y": 300})
    return ufo


@pytest.mark.parametrize(
    "input_expected",
    [
        ("top", (False, "top", None, False, False)),
        ("top_", (False, "top_", None, False, False)),
        ("top1", (False, "top1", None, False, False)),
        ("_bottom", (True, "bottom", None, False, False)),
        ("bottom_2", (False, "bottom", 2, False, False)),
        ("top_right_1", (False, "top_right", 1, False, False)),
    ],
)
def test_parseAnchorName(input_expected):
    anchorName, (isMark, key, number, isContextual, isIgnorable) = input_expected
    assert parseAnchorName(anchorName) == (
        isMark,
        key,
        number,
        isContextual,
        isIgnorable,
    )


def test_parseAnchorName_invalid():
    with pytest.raises(ValueError, match="mark anchor cannot be numbered"):
        parseAnchorName("_top_2")
    with pytest.raises(ValueError, match="mark anchor key is nil"):
        parseAnchorName("_")


def test_NamedAnchor_invalid():
    with pytest.raises(ValueError, match="indexes must start from 1"):
        NamedAnchor("top_0", 1, 2)


def test_NamedAnchor_repr():
    expected = "NamedAnchor(name='top', x=1.0, y=2.0)"
    assert repr(NamedAnchor("top", 1.0, 2.0)) == expected


class MarkFeatureWriterTest(FeatureWriterTest):
    FeatureWriter = MarkFeatureWriter

    def test__makeMarkClassDefinitions_empty(self, FontClass):
        ufo = FontClass()
        ufo.newGlyph("a").appendAnchor({"name": "top", "x": 250, "y": 500})
        ufo.newGlyph("c").appendAnchor({"name": "bottom", "x": 250, "y": -100})
        ufo.newGlyph("grave").appendAnchor({"name": "_top", "x": 100, "y": 200})
        ufo.newGlyph("cedilla").appendAnchor({"name": "_bottom", "x": 100, "y": 0})
        writer = MarkFeatureWriter()
        feaFile = ast.FeatureFile()
        writer.setContext(ufo, feaFile)
        markClassDefs = writer._makeMarkClassDefinitions()

        assert len(feaFile.markClasses) == 2
        assert [str(mcd) for mcd in markClassDefs] == [
            "markClass cedilla <anchor 100 0> @MC_bottom;",
            "markClass grave <anchor 100 200> @MC_top;",
        ]

    def test__makeMarkClassDefinitions_non_empty(self, FontClass):
        ufo = FontClass()
        ufo.newGlyph("a").appendAnchor({"name": "top", "x": 250, "y": 500})
        ufo.newGlyph("c").appendAnchor({"name": "bottom", "x": 250, "y": -100})
        ufo.newGlyph("grave").appendAnchor({"name": "_top", "x": 100, "y": 200})
        ufo.newGlyph("cedilla").appendAnchor({"name": "_bottom", "x": 100, "y": 0})
        ufo.features.text = dedent(
            """\
            markClass cedilla <anchor 200 0> @MC_bottom;
            markClass grave <anchor 100 200> @MC_top;
            """
        )

        writer = MarkFeatureWriter()
        feaFile = parseLayoutFeatures(ufo)
        writer.setContext(ufo, feaFile)
        markClassDefs = writer._makeMarkClassDefinitions()

        assert len(markClassDefs) == 1
        assert len(feaFile.markClasses) == 3
        assert "MC_bottom" in feaFile.markClasses
        assert "MC_top" in feaFile.markClasses
        assert [str(mcd) for mcd in markClassDefs] == [
            "markClass cedilla <anchor 100 0> @MC_bottom_1;"
        ]

    def test_skip_empty_feature(self, FontClass):
        ufo = FontClass()
        assert not self.writeFeatures(ufo)

        ufo.newGlyph("a").appendAnchor({"name": "top", "x": 100, "y": 200})
        ufo.newGlyph("acutecomb").appendAnchor({"name": "_top", "x": 100, "y": 200})

        fea = str(self.writeFeatures(ufo))

        assert "feature mark" in fea
        assert "feature mkmk" not in fea

    def test_skip_unnamed_anchors(self, FontClass, caplog):
        caplog.set_level(logging.ERROR)

        ufo = FontClass()
        ufo.newGlyph("a").appendAnchor({"x": 100, "y": 200})

        writer = MarkFeatureWriter()
        feaFile = ast.FeatureFile()

        logger = "ufo2ft.featureWriters.markFeatureWriter.MarkFeatureWriter"
        with caplog.at_level(logging.WARNING, logger=logger):
            writer.setContext(ufo, feaFile)

        assert len(caplog.records) == 1
        assert "unnamed anchor discarded in glyph 'a'" in caplog.text

    def test_warn_duplicate_anchor_names(self, FontClass, caplog):
        caplog.set_level(logging.ERROR)

        ufo = FontClass()
        ufo.newGlyph("a").anchors = [
            {"name": "top", "x": 100, "y": 200},
            {"name": "top", "x": 200, "y": 300},
        ]

        writer = MarkFeatureWriter()
        feaFile = ast.FeatureFile()

        logger = "ufo2ft.featureWriters.markFeatureWriter.MarkFeatureWriter"
        with caplog.at_level(logging.WARNING, logger=logger):
            writer.setContext(ufo, feaFile)

        assert len(caplog.records) == 1
        assert "duplicate anchor 'top' in glyph 'a'" in caplog.text

    def test_warn_liga_anchor_in_mark_glyph(self, testufo, caplog):
        caplog.set_level(logging.ERROR)

        testufo.newGlyph("ogonekcomb").anchors = [
            {"name": "_top", "x": 200, "y": -40},
            {"name": "top_1", "x": 200, "y": 450},  # should not be there!
        ]

        logger = "ufo2ft.featureWriters.markFeatureWriter.MarkFeatureWriter"
        with caplog.at_level(logging.WARNING, logger=logger):
            _ = self.writeFeatures(testufo)

        assert len(caplog.records) == 1
        assert "invalid ligature anchor 'top_1' in mark glyph" in caplog.text

    def test_ligature_NULL_anchor(self, testufo):
        testufo.newGlyph("f_f_foo").anchors = [
            {"name": "top_1", "x": 250, "y": 600},
            {"name": "top_2", "x": 500, "y": 600},
            {"name": "_3", "x": 0, "y": 0},  # this becomes <anchor NULL>
        ]
        generated = self.writeFeatures(testufo)

        assert re.search(r"ligComponent\s+<anchor NULL>", str(generated))

    def test_skip_existing_feature(self, testufo):
        testufo.features.text = dedent(
            """\
            markClass acutecomb <anchor 100 200> @MC_top;
            feature mark {
                lookup mark1 {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark1;
            } mark;
            """
        )

        generated = self.writeFeatures(testufo)

        # only mkmk is generated, mark was already present
        assert str(generated) == dedent(
            """\
            markClass tildecomb <anchor 100 200> @MC_top;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;
            """
        )

    def test_append_feature(self, testufo):
        testufo.features.text = dedent(
            """\
            markClass acutecomb <anchor 100 200> @MC_top;
            feature mark {
                lookup mark1 {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark1;
            } mark;
            """
        )

        generated = self.writeFeatures(testufo, mode="append")

        assert str(generated) == dedent(
            """\
            markClass tildecomb <anchor 100 200> @MC_top;

            feature mark {
                lookup mark2base {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark2base;

                lookup mark2liga {
                    pos ligature f_i
                            <anchor 100 500> mark @MC_top
                        ligComponent
                            <anchor 600 500> mark @MC_top;
                } mark2liga;

            } mark;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;
            """
        )

    def test_insert_comment_before(self, testufo):
        writer = MarkFeatureWriter()
        testufo.features.text = dedent(
            """\
            markClass acutecomb <anchor 100 200> @MC_top;
            feature mark {
                #
                # Automatic Code
                #
                lookup mark1 {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark1;
            } mark;
            """
        )
        feaFile = parseLayoutFeatures(testufo)

        assert writer.write(testufo, feaFile)

        assert str(feaFile) == dedent(
            """\
            markClass tildecomb <anchor 100 200> @MC_top;

            markClass acutecomb <anchor 100 200> @MC_top;
            feature mark {
                lookup mark2base {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark2base;

                lookup mark2liga {
                    pos ligature f_i
                            <anchor 100 500> mark @MC_top
                        ligComponent
                            <anchor 600 500> mark @MC_top;
                } mark2liga;

            } mark;

            feature mark {
                #
                #
                lookup mark1 {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark1;

            } mark;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;
            """
        )

        # test append mode ignores insert marker
        generated = self.writeFeatures(testufo, mode="append")
        assert str(generated) == dedent(
            """\
            markClass tildecomb <anchor 100 200> @MC_top;

            feature mark {
                lookup mark2base {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark2base;

                lookup mark2liga {
                    pos ligature f_i
                            <anchor 100 500> mark @MC_top
                        ligComponent
                            <anchor 600 500> mark @MC_top;
                } mark2liga;

            } mark;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;
            """
        )

    def test_insert_comment_after(self, testufo):
        writer = MarkFeatureWriter()
        testufo.features.text = dedent(
            """\
            markClass acutecomb <anchor 100 200> @MC_top;
            feature mark {
                lookup mark1 {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark1;
                #
                # Automatic Code
                #
            } mark;
            """
        )
        feaFile = parseLayoutFeatures(testufo)

        assert writer.write(testufo, feaFile)

        assert str(feaFile) == dedent(
            """\
            markClass tildecomb <anchor 100 200> @MC_top;

            markClass acutecomb <anchor 100 200> @MC_top;
            feature mark {
                lookup mark1 {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark1;

                #
                #
            } mark;

            feature mark {
                lookup mark2base {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark2base;

                lookup mark2liga {
                    pos ligature f_i
                            <anchor 100 500> mark @MC_top
                        ligComponent
                            <anchor 600 500> mark @MC_top;
                } mark2liga;

            } mark;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;
            """
        )

        # test append mode ignores insert marker
        generated = self.writeFeatures(testufo, mode="append")
        assert str(generated) == dedent(
            """\
            markClass tildecomb <anchor 100 200> @MC_top;

            feature mark {
                lookup mark2base {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark2base;

                lookup mark2liga {
                    pos ligature f_i
                            <anchor 100 500> mark @MC_top
                        ligComponent
                            <anchor 600 500> mark @MC_top;
                } mark2liga;

            } mark;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;
            """
        )

    def test_insert_comment_middle(self, testufo):
        writer = MarkFeatureWriter()
        testufo.features.text = dedent(
            """\
            markClass acutecomb <anchor 100 200> @MC_top;
            feature mark {
                lookup mark1 {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark1;
                #
                # Automatic Code
                #
                lookup mark2 {
                    pos base a
                        <anchor 150 250> mark @MC_top;
                } mark2;
            } mark;
            """
        )
        feaFile = parseLayoutFeatures(testufo)

        writer.write(testufo, feaFile)
        assert str(feaFile) == dedent(
            """\
            markClass tildecomb <anchor 100 200> @MC_top;

            markClass acutecomb <anchor 100 200> @MC_top;
            feature mark {
                lookup mark1 {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark1;

                #
            } mark;

            feature mark {
                lookup mark2base {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark2base;

                lookup mark2liga {
                    pos ligature f_i
                            <anchor 100 500> mark @MC_top
                        ligComponent
                            <anchor 600 500> mark @MC_top;
                } mark2liga;

            } mark;

            feature mark {
                #
                lookup mark2 {
                    pos base a
                        <anchor 150 250> mark @MC_top;
                } mark2;

            } mark;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;
            """
        )

        # test append mode ignores insert marker
        generated = self.writeFeatures(testufo, mode="append")
        assert str(generated) == dedent(
            """\
            markClass tildecomb <anchor 100 200> @MC_top;

            feature mark {
                lookup mark2base {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark2base;

                lookup mark2liga {
                    pos ligature f_i
                            <anchor 100 500> mark @MC_top
                        ligComponent
                            <anchor 600 500> mark @MC_top;
                } mark2liga;

            } mark;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;
            """
        )

    def test_insert_comment_outside_block(self, testufo):
        writer = MarkFeatureWriter()
        testufo.features.text = dedent(
            """\
            #
            # Automatic Code
            #
            """
        )
        feaFile = parseLayoutFeatures(testufo)

        assert writer.write(testufo, feaFile)

        testufo.features.text = dedent(
            """\
            #
            # Automatic Code
            #
            markClass acutecomb <anchor 100 200> @MC_top;
            feature mark {
                lookup mark1 {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark1;

            } mark;
            """
        )
        feaFile = parseLayoutFeatures(testufo)

        assert writer.write(testufo, feaFile)

        # test append mode
        writer = MarkFeatureWriter(mode="append")
        assert writer.write(testufo, feaFile)

    def test_defs_and_lookups_first(self, testufo):
        testufo.newGlyph("circumflexcomb")
        writer = MarkFeatureWriter()
        testufo.features.text = dedent(
            """\
            feature mkmk {
                # Automatic Code
                # Move acutecomb down and right if preceded by circumflexcomb
                lookup move_acutecomb {
                    lookupflag UseMarkFilteringSet [acutecomb circumflexcomb];
                    pos circumflexcomb acutecomb' <0 20 0 20>;
                } move_acutecomb;
            } mkmk;
            """
        )
        feaFile = parseLayoutFeatures(testufo)

        assert writer.write(testufo, feaFile)

        assert str(feaFile) == dedent(
            """\
            markClass acutecomb <anchor 100 200> @MC_top;
            markClass tildecomb <anchor 100 200> @MC_top;

            feature mark {
                lookup mark2base {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark2base;

                lookup mark2liga {
                    pos ligature f_i
                            <anchor 100 500> mark @MC_top
                        ligComponent
                            <anchor 600 500> mark @MC_top;
                } mark2liga;

            } mark;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;

            feature mkmk {
                # Move acutecomb down and right if preceded by circumflexcomb
                lookup move_acutecomb {
                    lookupflag UseMarkFilteringSet [acutecomb circumflexcomb];
                    pos circumflexcomb acutecomb' <0 20 0 20>;
                } move_acutecomb;

            } mkmk;
            """
        )

    def test_mark_mkmk_features(self, testufo):
        writer = MarkFeatureWriter()  # by default both mark + mkmk are built
        feaFile = ast.FeatureFile()
        assert writer.write(testufo, feaFile)

        assert str(feaFile) == dedent(
            """\
            markClass acutecomb <anchor 100 200> @MC_top;
            markClass tildecomb <anchor 100 200> @MC_top;

            feature mark {
                lookup mark2base {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark2base;

                lookup mark2liga {
                    pos ligature f_i
                            <anchor 100 500> mark @MC_top
                        ligComponent
                            <anchor 600 500> mark @MC_top;
                } mark2liga;

            } mark;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;
            """
        )

    def test_write_only_one(self, testufo):
        writer = MarkFeatureWriter(features=["mkmk"])  # only builds "mkmk"
        feaFile = ast.FeatureFile()
        assert writer.write(testufo, feaFile)
        fea = str(feaFile)

        assert "feature mark" not in fea
        assert "feature mkmk" in fea

        writer = MarkFeatureWriter(features=["mark"])  # only builds "mark"
        feaFile = ast.FeatureFile()
        assert writer.write(testufo, feaFile)
        fea = str(feaFile)

        assert "feature mark" in fea
        assert "feature mkmk" not in fea

    def test_predefined_anchor_lists(self, FontClass):
        """Roboto uses some weird anchor naming scheme, see:
        https://github.com/google/roboto/blob/
            5700de83856781fa0c097a349e46dbaae5792cb0/
            scripts/lib/fontbuild/markFeature.py#L41-L47
        """

        class RobotoMarkFeatureWriter(MarkFeatureWriter):
            class NamedAnchor(NamedAnchor):
                markPrefix = "_mark"
                ignoreRE = "(^mkmk|_acc$)"

        ufo = FontClass()
        a = ufo.newGlyph("a")
        a.anchors = [
            {"name": "top", "x": 250, "y": 600},
            {"name": "bottom", "x": 250, "y": -100},
        ]
        f_i = ufo.newGlyph("f_i")
        f_i.anchors = [
            {"name": "top_1", "x": 200, "y": 700},
            {"name": "top_2", "x": 500, "y": 700},
        ]
        gravecomb = ufo.newGlyph("gravecomb")
        gravecomb.anchors = [
            {"name": "_marktop", "x": 160, "y": 780},
            {"name": "mkmktop", "x": 150, "y": 800},
            {"name": "mkmkbottom_acc", "x": 150, "y": 600},
        ]
        ufo.newGlyph("cedillacomb").appendAnchor(
            {"name": "_markbottom", "x": 200, "y": 0}
        )
        ufo.newGlyph("ogonekcomb").appendAnchor({"name": "_bottom", "x": 180, "y": -10})

        writer = RobotoMarkFeatureWriter(groupMarkClasses=True)
        feaFile = ast.FeatureFile()
        writer.write(ufo, feaFile)

        assert str(feaFile) == dedent(
            """\
            markClass cedillacomb <anchor 200 0> @MC_markbottom;
            markClass gravecomb <anchor 160 780> @MC_marktop;

            feature mark {
                lookup mark2base {
                    pos base a
                        <anchor 250 -100> mark @MC_markbottom
                        <anchor 250 600> mark @MC_marktop;
                } mark2base;

                lookup mark2liga {
                    pos ligature f_i
                            <anchor 200 700> mark @MC_marktop
                        ligComponent
                            <anchor 500 700> mark @MC_marktop;
                } mark2liga;

            } mark;

            feature mkmk {
                lookup mark2mark_bottom {
                    @MFS_mark2mark_bottom = [cedillacomb gravecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_bottom;
                    pos mark gravecomb
                        <anchor 150 600> mark @MC_markbottom;
                } mark2mark_bottom;

                lookup mark2mark_top {
                    @MFS_mark2mark_top = [gravecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark gravecomb
                        <anchor 150 800> mark @MC_marktop;
                } mark2mark_top;

            } mkmk;
            """  # noqa: B950
        )

    @pytest.mark.parametrize(
        "groupMarkClasses, expected",
        [
            (
                True,
                dedent(
                    """\
                    markClass nukta-kannada <anchor 0 0> @MC_bottom;
                    markClass candrabindu-kannada <anchor 0 547> @MC_top;
                    markClass halant-kannada <anchor -456 460> @MC_topright;

                    feature abvm {
                        lookup abvm_mark2base {
                            pos base ka-kannada.base
                                <anchor 291 547> mark @MC_top
                                <anchor 391 460> mark @MC_topright;
                        } abvm_mark2base;

                    } abvm;

                    feature blwm {
                        lookup blwm_mark2base {
                            pos base ka-kannada
                                <anchor 290 0> mark @MC_bottom;
                            pos base ka-kannada.base
                                <anchor 290 0> mark @MC_bottom;
                        } blwm_mark2base;

                    } blwm;

                    feature mark {
                        lookup mark2base {
                            pos base dottedCircle
                                <anchor 297 0> mark @MC_bottom
                                <anchor 297 552> mark @MC_top
                                <anchor 491 458> mark @MC_topright;
                        } mark2base;

                    } mark;
                    """  # noqa: B950
                ),
            ),
            (
                False,
                dedent(
                    """\
                    markClass nukta-kannada <anchor 0 0> @MC_bottom;
                    markClass candrabindu-kannada <anchor 0 547> @MC_top;
                    markClass halant-kannada <anchor -456 460> @MC_topright;

                    feature abvm {
                        lookup abvm_mark2base {
                            pos base ka-kannada.base
                                <anchor 291 547> mark @MC_top;
                        } abvm_mark2base;

                        lookup abvm_mark2base_1 {
                            pos base ka-kannada.base
                                <anchor 391 460> mark @MC_topright;
                        } abvm_mark2base_1;

                    } abvm;

                    feature blwm {
                        lookup blwm_mark2base {
                            pos base ka-kannada
                                <anchor 290 0> mark @MC_bottom;
                            pos base ka-kannada.base
                                <anchor 290 0> mark @MC_bottom;
                        } blwm_mark2base;

                    } blwm;

                    feature mark {
                        lookup mark2base {
                            pos base dottedCircle
                                <anchor 297 0> mark @MC_bottom;
                        } mark2base;

                        lookup mark2base_1 {
                            pos base dottedCircle
                                <anchor 297 552> mark @MC_top;
                        } mark2base_1;

                        lookup mark2base_2 {
                            pos base dottedCircle
                                <anchor 491 458> mark @MC_topright;
                        } mark2base_2;

                    } mark;
                    """  # noqa: B950
                ),
            ),
        ],
    )
    def test_abvm_blwm_features(self, FontClass, groupMarkClasses, expected):
        ufo = FontClass()
        ufo.info.unitsPerEm = 1000

        dottedCircle = ufo.newGlyph("dottedCircle")
        dottedCircle.unicode = 0x25CC
        dottedCircle.anchors = [
            {"name": "top", "x": 297, "y": 552},
            {"name": "topright", "x": 491, "y": 458},
            {"name": "bottom", "x": 297, "y": 0},
        ]

        nukta = ufo.newGlyph("nukta-kannada")
        nukta.unicode = 0x0CBC
        nukta.appendAnchor({"name": "_bottom", "x": 0, "y": 0})

        nukta = ufo.newGlyph("candrabindu-kannada")
        nukta.unicode = 0x0C81
        nukta.appendAnchor({"name": "_top", "x": 0, "y": 547})

        halant = ufo.newGlyph("halant-kannada")
        halant.unicode = 0x0CCD
        halant.appendAnchor({"name": "_topright", "x": -456, "y": 460})

        ka = ufo.newGlyph("ka-kannada")
        ka.unicode = 0x0C95
        ka.appendAnchor({"name": "bottom", "x": 290, "y": 0})

        ka_base = ufo.newGlyph("ka-kannada.base")
        ka_base.appendAnchor({"name": "top", "x": 291, "y": 547})
        ka_base.appendAnchor({"name": "topright", "x": 391, "y": 460})
        ka_base.appendAnchor({"name": "bottom", "x": 290, "y": 0})

        ufo.features.text = dedent(
            """\
            languagesystem DFLT dflt;
            languagesystem knda dflt;
            languagesystem knd2 dflt;

            feature psts {
                sub ka-kannada' halant-kannada by ka-kannada.base;
            } psts;
            """
        )
        generated = self.writeFeatures(ufo, groupMarkClasses=groupMarkClasses)

        assert str(generated) == expected

    def test_contextual_abvm_blwm_anchors(self, FontClass):
        ufo = FontClass()
        ufo.info.unitsPerEm = 1000

        iMark = ufo.newGlyph("iMark-khmer")
        iMark.unicode = 0x17B7
        iMark.appendAnchor({"name": "_topright", "x": 412, "y": 600})

        iiMark = ufo.newGlyph("iiMark-khmer")
        iiMark.unicode = 0x17B8
        iiMark.appendAnchor({"name": "_topright", "x": 412, "y": 600})

        kaBelow = ufo.newGlyph("ka-khmer.below")
        kaBelow.appendAnchor({"name": "_bottom", "x": 276, "y": 0})

        ka = ufo.newGlyph("ka-khmer")
        ka.unicode = 0x1780
        ka.appendAnchor({"name": "topright", "x": 470, "y": 600})
        ka.appendAnchor(
            {"name": "*topright", "x": 276, "y": 700, "identifier": "*topright"}
        )
        ka.appendAnchor({"name": "bottom", "x": 276, "y": 0})
        ka.appendAnchor(
            {"name": "*bottom", "x": 276, "y": 100, "identifier": "*bottom"}
        )

        ka.lib[OBJECT_LIBS_KEY] = {
            "*topright": {
                "GPOS_Context": "lookupflag UseMarkFilteringSet [iMark-khmer]; "
                "* ka-khmer iMark-khmer"
            },
            "*bottom": {
                "GPOS_Context": "lookupflag UseMarkFilteringSet [ka-khmer.below]; *"
            },
        }

        writer = MarkFeatureWriter()
        feaFile = ast.FeatureFile()
        assert str(feaFile) == ""
        assert writer.write(ufo, feaFile)

        assert str(feaFile) == dedent(
            """\
            markClass ka-khmer.below <anchor 276 0> @MC_bottom;
            markClass iMark-khmer <anchor 412 600> @MC_topright;
            markClass iiMark-khmer <anchor 412 600> @MC_topright;

            lookup abvm_mark2base {
                pos base ka-khmer
                    <anchor 470 600> mark @MC_topright;
            } abvm_mark2base;

            lookup ContextualAbvm_0 {
                pos base ka-khmer
                    <anchor 276 700> mark @MC_topright;
            } ContextualAbvm_0;

            lookup ContextualAbvmDispatch_0 {
                lookupflag UseMarkFilteringSet [iMark-khmer];
                # * ka-khmer iMark-khmer
                pos [ka-khmer] @MC_topright' lookup ContextualAbvm_0 ka-khmer iMark-khmer;
            } ContextualAbvmDispatch_0;

            lookup blwm_mark2base {
                pos base ka-khmer
                    <anchor 276 0> mark @MC_bottom;
            } blwm_mark2base;

            lookup ContextualBlwm_0 {
                pos base ka-khmer
                    <anchor 276 100> mark @MC_bottom;
            } ContextualBlwm_0;

            lookup ContextualBlwmDispatch_0 {
                lookupflag UseMarkFilteringSet [ka-khmer.below];
                # *
                pos [ka-khmer] @MC_bottom' lookup ContextualBlwm_0;
            } ContextualBlwmDispatch_0;

            feature abvm {
                lookup abvm_mark2base;
                lookup ContextualAbvmDispatch_0;
            } abvm;

            feature blwm {
                lookup blwm_mark2base;
                lookup ContextualBlwmDispatch_0;
            } blwm;
            """
        )

    def test_shared_script_char(self, FontClass):
        ufo = FontClass()
        ufo.info.unitsPerEm = 1000

        dottedCircle = ufo.newGlyph("kashida-ar")
        dottedCircle.unicode = 0x0640
        dottedCircle.anchors = [
            {"name": "top", "x": 100, "y": 100},
            {"name": "bottom", "x": 100, "y": -100},
        ]

        nukta = ufo.newGlyph("fatha-ar")
        nukta.unicode = 0x064E
        nukta.appendAnchor({"name": "_top", "x": 0, "y": 0})

        nukta = ufo.newGlyph("kasra-ar")
        nukta.unicode = 0x0650
        nukta.appendAnchor({"name": "_bottom", "x": 0, "y": 547})

        ufo.features.text = dedent(
            """\
            languagesystem DFLT dflt;
            languagesystem arab dflt;
            """
        )
        generated = self.writeFeatures(ufo, groupMarkClasses=True)

        assert str(generated) == dedent(
            """\
            markClass kasra-ar <anchor 0 547> @MC_bottom;
            markClass fatha-ar <anchor 0 0> @MC_top;

            feature mark {
                lookup mark2base {
                    pos base kashida-ar
                        <anchor 100 -100> mark @MC_bottom
                        <anchor 100 100> mark @MC_top;
                } mark2base;

            } mark;
            """  # noqa: B950
        )

        expected = dedent(
            """\
            markClass kasra-ar <anchor 0 547> @MC_bottom;
            markClass fatha-ar <anchor 0 0> @MC_top;

            feature abvm {
                lookup abvm_mark2base {
                    pos base kashida-ar
                        <anchor 100 100> mark @MC_top;
                } abvm_mark2base;

            } abvm;

            feature blwm {
                lookup blwm_mark2base {
                    pos base kashida-ar
                        <anchor 100 -100> mark @MC_bottom;
                } blwm_mark2base;

            } blwm;

            feature mark {
                lookup mark2base {
                    pos base kashida-ar
                        <anchor 100 -100> mark @MC_bottom
                        <anchor 100 100> mark @MC_top;
                } mark2base;

            } mark;
            """  # noqa: B950
        )

        ufo.features.text = ""
        generated = self.writeFeatures(ufo, groupMarkClasses=True)
        assert str(generated) == expected

        ufo.features.text = dedent(
            """\
            languagesystem DFLT dflt;
            languagesystem arab dflt;
            languagesystem adlm dflt;
            """
        )
        generated = self.writeFeatures(ufo, groupMarkClasses=True)
        assert str(generated) == expected

    @pytest.mark.parametrize(
        "groupMarkClasses, expected",
        [
            (
                True,
                dedent(
                    """\
                    markClass barcomb <anchor 100 40> @MC_bar;
                    markClass cedillacomb <anchor 10 -5> @MC_bottom;
                    markClass bazcomb <anchor 90 320> @MC_bottom;
                    markClass acutecomb <anchor 100 200> @MC_top;
                    markClass tildecomb <anchor 100 200> @MC_top;
                    markClass foocomb <anchor 100 40> @MC_top;

                    feature abvm {
                        lookup abvm_mark2liga {
                            pos ligature foo_bar_baz
                                    <anchor 100 500> mark @MC_top
                                ligComponent
                                    <anchor NULL>
                                ligComponent
                                    <anchor 1100 499> mark @MC_bar
                                    <anchor 1000 500> mark @MC_top;
                            pos ligature bar_foo
                                    <anchor NULL>
                                ligComponent
                                    <anchor 600 501> mark @MC_top;
                        } abvm_mark2liga;

                        lookup abvm_mark2mark_bar {
                            @MFS_abvm_mark2mark_bar = [barcomb];
                            lookupflag UseMarkFilteringSet @MFS_abvm_mark2mark_bar;
                            pos mark barcomb
                                <anchor 100 440> mark @MC_bar;
                        } abvm_mark2mark_bar;

                        lookup abvm_mark2mark_top {
                            @MFS_abvm_mark2mark_top = [foocomb];
                            lookupflag UseMarkFilteringSet @MFS_abvm_mark2mark_top;
                            pos mark foocomb
                                <anchor 100 190> mark @MC_top;
                        } abvm_mark2mark_top;

                    } abvm;

                    feature blwm {
                        lookup blwm_mark2liga {
                            pos ligature foo_bar_baz
                                    <anchor 100 10> mark @MC_bottom
                                ligComponent
                                    <anchor NULL>
                                ligComponent
                                    <anchor NULL>;
                        } blwm_mark2liga;

                        lookup blwm_mark2mark_bottom {
                            @MFS_blwm_mark2mark_bottom = [bazcomb];
                            lookupflag UseMarkFilteringSet @MFS_blwm_mark2mark_bottom;
                            pos mark bazcomb
                                <anchor 100 -34> mark @MC_bottom;
                        } blwm_mark2mark_bottom;

                    } blwm;

                    feature mark {
                        lookup mark2base {
                            pos base a
                                <anchor 100 200> mark @MC_top;
                            pos base c
                                <anchor 240 0> mark @MC_bottom;
                            pos base dottedCircle
                                <anchor 491 458> mark @MC_bar
                                <anchor 297 0> mark @MC_bottom
                                <anchor 297 552> mark @MC_top;
                        } mark2base;

                        lookup mark2liga {
                            pos ligature f_i
                                    <anchor 100 500> mark @MC_top
                                ligComponent
                                    <anchor 600 500> mark @MC_top;
                        } mark2liga;

                    } mark;

                    feature mkmk {
                        lookup mark2mark_bottom {
                            @MFS_mark2mark_bottom = [cedillacomb];
                            lookupflag UseMarkFilteringSet @MFS_mark2mark_bottom;
                            pos mark cedillacomb
                                <anchor 20 -309> mark @MC_bottom;
                        } mark2mark_bottom;

                        lookup mark2mark_top {
                            @MFS_mark2mark_top = [acutecomb tildecomb];
                            lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                            pos mark tildecomb
                                <anchor 100 300> mark @MC_top;
                        } mark2mark_top;

                    } mkmk;
                    """  # noqa: B950
                ),
            ),
            (
                False,
                dedent(
                    """\
                    markClass barcomb <anchor 100 40> @MC_bar;
                    markClass cedillacomb <anchor 10 -5> @MC_bottom;
                    markClass bazcomb <anchor 90 320> @MC_bottom;
                    markClass acutecomb <anchor 100 200> @MC_top;
                    markClass tildecomb <anchor 100 200> @MC_top;
                    markClass foocomb <anchor 100 40> @MC_top;

                    feature abvm {
                        lookup abvm_mark2liga {
                            pos ligature foo_bar_baz
                                    <anchor NULL>
                                ligComponent
                                    <anchor NULL>
                                ligComponent
                                    <anchor 1100 499> mark @MC_bar;
                        } abvm_mark2liga;

                        lookup abvm_mark2liga_1 {
                            pos ligature foo_bar_baz
                                    <anchor 100 500> mark @MC_top
                                ligComponent
                                    <anchor NULL>
                                ligComponent
                                    <anchor 1000 500> mark @MC_top;
                            pos ligature bar_foo
                                    <anchor NULL>
                                ligComponent
                                    <anchor 600 501> mark @MC_top;
                        } abvm_mark2liga_1;

                        lookup abvm_mark2mark_bar {
                            @MFS_abvm_mark2mark_bar = [barcomb];
                            lookupflag UseMarkFilteringSet @MFS_abvm_mark2mark_bar;
                            pos mark barcomb
                                <anchor 100 440> mark @MC_bar;
                        } abvm_mark2mark_bar;

                        lookup abvm_mark2mark_top {
                            @MFS_abvm_mark2mark_top = [foocomb];
                            lookupflag UseMarkFilteringSet @MFS_abvm_mark2mark_top;
                            pos mark foocomb
                                <anchor 100 190> mark @MC_top;
                        } abvm_mark2mark_top;

                    } abvm;

                    feature blwm {
                        lookup blwm_mark2liga {
                            pos ligature foo_bar_baz
                                    <anchor 100 10> mark @MC_bottom
                                ligComponent
                                    <anchor NULL>
                                ligComponent
                                    <anchor NULL>;
                        } blwm_mark2liga;

                        lookup blwm_mark2mark_bottom {
                            @MFS_blwm_mark2mark_bottom = [bazcomb];
                            lookupflag UseMarkFilteringSet @MFS_blwm_mark2mark_bottom;
                            pos mark bazcomb
                                <anchor 100 -34> mark @MC_bottom;
                        } blwm_mark2mark_bottom;

                    } blwm;

                    feature mark {
                        lookup mark2base {
                            pos base dottedCircle
                                <anchor 491 458> mark @MC_bar;
                        } mark2base;

                        lookup mark2base_1 {
                            pos base c
                                <anchor 240 0> mark @MC_bottom;
                            pos base dottedCircle
                                <anchor 297 0> mark @MC_bottom;
                        } mark2base_1;

                        lookup mark2base_2 {
                            pos base a
                                <anchor 100 200> mark @MC_top;
                            pos base dottedCircle
                                <anchor 297 552> mark @MC_top;
                        } mark2base_2;

                        lookup mark2liga {
                            pos ligature f_i
                                    <anchor 100 500> mark @MC_top
                                ligComponent
                                    <anchor 600 500> mark @MC_top;
                        } mark2liga;

                    } mark;

                    feature mkmk {
                        lookup mark2mark_bottom {
                            @MFS_mark2mark_bottom = [cedillacomb];
                            lookupflag UseMarkFilteringSet @MFS_mark2mark_bottom;
                            pos mark cedillacomb
                                <anchor 20 -309> mark @MC_bottom;
                        } mark2mark_bottom;

                        lookup mark2mark_top {
                            @MFS_mark2mark_top = [acutecomb tildecomb];
                            lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                            pos mark tildecomb
                                <anchor 100 300> mark @MC_top;
                        } mark2mark_top;

                    } mkmk;
                    """  # noqa: B950
                ),
            ),
        ],
    )
    def test_all_features(self, testufo, groupMarkClasses, expected):
        ufo = testufo
        ufo.info.unitsPerEm = 1000

        ufo.newGlyph("cedillacomb").anchors = [
            {"name": "_bottom", "x": 10, "y": -5},
            {"name": "bottom", "x": 20, "y": -309},
        ]
        ufo.newGlyph("c").appendAnchor({"name": "bottom", "x": 240, "y": 0})

        dottedCircle = ufo.newGlyph("dottedCircle")
        dottedCircle.unicode = 0x25CC
        dottedCircle.anchors = [
            {"name": "top", "x": 297, "y": 552},
            {"name": "bottom", "x": 297, "y": 0},
            {"name": "bar", "x": 491, "y": 458},
        ]

        # too lazy, couldn't come up with a real-word example :/
        foocomb = ufo.newGlyph("foocomb")
        foocomb.unicode = 0x0B85
        foocomb.anchors = [
            {"name": "_top", "x": 100, "y": 40},
            {"name": "top", "x": 100, "y": 190},
        ]
        barcomb = ufo.newGlyph("barcomb")
        barcomb.unicode = 0x0B86
        barcomb.anchors = [
            {"name": "_bar", "x": 100, "y": 40},
            {"name": "bar", "x": 100, "y": 440.1},
        ]
        bazcomb = ufo.newGlyph("bazcomb")
        bazcomb.unicode = 0x0B87
        bazcomb.anchors = [
            {"name": "_bottom", "x": 90, "y": 320},
            {"name": "bottom", "x": 100, "y": -34},
        ]
        foo_bar_baz = ufo.newGlyph("foo_bar_baz")
        foo_bar_baz.unicode = 0x0B88
        foo_bar_baz.anchors = [
            {"name": "top_1", "x": 100, "y": 500},
            {"name": "bottom_1", "x": 100, "y": 10},
            {"name": "_2", "x": 600, "y": 500},
            {"name": "top_3", "x": 1000, "y": 500},
            {"name": "bar_3", "x": 1100, "y": 499},  # below half UPEM
        ]
        bar_foo = ufo.newGlyph("bar_foo")
        bar_foo.unicode = 0x0B89
        # sequence doesn't start from 1, the first is implied NULL anchor
        bar_foo.anchors = [{"name": "top_2", "x": 600, "y": 501}]

        testufo.glyphOrder = [
            "a",
            "f_i",
            "acutecomb",
            "tildecomb",
            "cedillacomb",
            "c",
            "dottedCircle",
            "foocomb",
            "barcomb",
            "bazcomb",
            "foo_bar_baz",
            "bar_foo",
        ]
        ufo.features.text = dedent(
            """\
            languagesystem DFLT dflt;
            languagesystem taml dflt;
            """
        )
        generated = self.writeFeatures(testufo, groupMarkClasses=groupMarkClasses)

        assert str(generated) == expected

    def test_mark_mkmk_features_with_GDEF(self, testufo):
        D = testufo.newGlyph("D")
        D.anchors = [
            {"name": "top", "x": 300, "y": 700},
            {"name": "center", "x": 320, "y": 360},
        ]
        # these glyphs have compatible anchors but since they not listed in
        # the GDEF groups, they won't be included in the mark/mkmk feature
        testufo.newGlyph("Alpha").appendAnchor({"name": "topleft", "x": -10, "y": 400})
        testufo.newGlyph("psili").appendAnchor({"name": "_topleft", "x": 0, "y": 50})
        dotaccentcomb = testufo.newGlyph("dotaccentcomb")
        # this mark glyph has more than one mark anchor, and both will be
        # generated. Since the two mark anchors cannot cohabit in the same
        # mark lookup, two lookups will be generated.
        dotaccentcomb.anchors = [
            {"name": "_center", "x": 0, "y": 0},
            {"name": "_top", "x": 0, "y": 0},
            {"name": "top", "x": 0, "y": 300},
        ]
        testufo.features.text = dedent(
            """\
            @Bases = [a D];
            @Marks = [acutecomb tildecomb dotaccentcomb];
            table GDEF {
                GlyphClassDef @Bases, [f_i], @Marks, ;
            } GDEF;
            """
        )
        testufo.glyphOrder = [
            "Alpha",
            "D",
            "a",
            "acutecomb",
            "dotaccentcomb",
            "f_i",
            "psili",
            "tildecomb",
        ]

        generated = self.writeFeatures(testufo)

        assert str(generated) == dedent(
            """\
            markClass dotaccentcomb <anchor 0 0> @MC_center;
            markClass acutecomb <anchor 100 200> @MC_top;
            markClass dotaccentcomb <anchor 0 0> @MC_top;
            markClass tildecomb <anchor 100 200> @MC_top;

            feature mark {
                lookup mark2base {
                    pos base D
                        <anchor 320 360> mark @MC_center;
                } mark2base;

                lookup mark2base_1 {
                    pos base D
                        <anchor 300 700> mark @MC_top;
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark2base_1;

                lookup mark2liga {
                    pos ligature f_i
                            <anchor 100 500> mark @MC_top
                        ligComponent
                            <anchor 600 500> mark @MC_top;
                } mark2liga;

            } mark;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb dotaccentcomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark dotaccentcomb
                        <anchor 0 300> mark @MC_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;
            """
        )

    def test_mark_mkmk_features_with_GDEF_and_openTypeCategories(self, testufo):
        # this glyph has compatible anchors and has an openTypeCategories "base"
        # value
        D = testufo.newGlyph("D")
        D.anchors = [
            {"name": "top", "x": 300, "y": 700},
            {"name": "center", "x": 320, "y": 360},
        ]
        # these glyphs have compatible anchors but since they not listed in
        # the GDEF groups, they won't be included in the mark/mkmk feature
        testufo.newGlyph("Alpha").appendAnchor({"name": "topleft", "x": -10, "y": 400})
        testufo.newGlyph("psili").appendAnchor({"name": "_topleft", "x": 0, "y": 50})
        dotaccentcomb = testufo.newGlyph("dotaccentcomb")
        # this mark glyph has more than one mark anchor, and both will be
        # generated. Since the two mark anchors cannot cohabit in the same
        # mark lookup, two lookups will be generated.
        dotaccentcomb.anchors = [
            {"name": "_center", "x": 0, "y": 0},
            {"name": "_top", "x": 0, "y": 0},
            {"name": "top", "x": 0, "y": 300},
        ]
        # will be ignored because in GDEF table below
        testufo.lib["public.openTypeCategories"] = {
            "D": "base",
            "dotaccentcomb": "mark",
            "tildecomb": "base",
        }
        testufo.features.text = dedent(
            """\
            @Bases = [a];
            @Marks = [acutecomb tildecomb];
            table GDEF {
                GlyphClassDef @Bases, [f_i], @Marks, ;
            } GDEF;
            """
        )
        testufo.glyphOrder = [
            "Alpha",
            "D",
            "a",
            "acutecomb",
            "dotaccentcomb",
            "f_i",
            "psili",
            "tildecomb",
        ]

        generated = self.writeFeatures(testufo)

        assert str(generated) == dedent(
            """\
            markClass acutecomb <anchor 100 200> @MC_top;
            markClass tildecomb <anchor 100 200> @MC_top;

            feature mark {
                lookup mark2base {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark2base;

                lookup mark2liga {
                    pos ligature f_i
                            <anchor 100 500> mark @MC_top
                        ligComponent
                            <anchor 600 500> mark @MC_top;
                } mark2liga;

            } mark;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;
            """
        )

    def test_multiple_anchor_classes_base(self, FontClass):
        dirname = os.path.dirname(os.path.dirname(__file__))
        fontPath = os.path.join(dirname, "data", "MultipleAnchorClasses.ufo")
        testufo = FontClass(fontPath)
        generated = self.writeFeatures(testufo)

        assert str(generated) == dedent(
            """\
            markClass acutecomb <anchor -175 589> @MC_topA;
            markClass acutecomb <anchor -175 572> @MC_topE;

            feature mark {
                lookup mark2base {
                    pos base a
                        <anchor 515 581> mark @MC_topA;
                } mark2base;

                lookup mark2base_1 {
                    pos base e
                        <anchor -21 396> mark @MC_topE;
                } mark2base_1;

            } mark;
            """
        )

    @pytest.mark.parametrize(
        "groupMarkClasses, expected",
        [
            (
                True,
                # 'MC_top' should be last thanks to the anchorSortKey. Arguably
                # this is wrong and might not match current Glyphs.app's behavior
                # but we decided to keep for backward compatibility with existing
                # projects.
                dedent(
                    """\
                    markClass acutecomb <anchor 100 200> @MC_top;
                    markClass acutecomb <anchor 150 250> @MC_topOther;

                    feature mark {
                        lookup mark2liga {
                            pos ligature f_f
                                    <anchor 101 501> mark @MC_topOther
                                ligComponent
                                    <anchor 601 501> mark @MC_topOther;
                            pos ligature f_l
                                    <anchor NULL>
                                ligComponent
                                    <anchor 602 502> mark @MC_topOther;
                        } mark2liga;

                        lookup mark2liga_1 {
                            pos ligature f_i
                                    <anchor 100 500> mark @MC_top
                                ligComponent
                                    <anchor 600 500> mark @MC_top;
                            pos ligature f_l
                                    <anchor 102 502> mark @MC_top
                                ligComponent
                                    <anchor NULL>;
                        } mark2liga_1;

                    } mark;
                    """
                ),
            ),
            (
                False,
                # with groupMarkClasses=False, lookups are simply sorted by mark
                # class name alphabetically so 'MC_topOther' is last and wins
                dedent(
                    """\
                    markClass acutecomb <anchor 100 200> @MC_top;
                    markClass acutecomb <anchor 150 250> @MC_topOther;

                    feature mark {
                        lookup mark2liga {
                            pos ligature f_i
                                    <anchor 100 500> mark @MC_top
                                ligComponent
                                    <anchor 600 500> mark @MC_top;
                            pos ligature f_l
                                    <anchor 102 502> mark @MC_top
                                ligComponent
                                    <anchor NULL>;
                        } mark2liga;

                        lookup mark2liga_1 {
                            pos ligature f_f
                                    <anchor 101 501> mark @MC_topOther
                                ligComponent
                                    <anchor 601 501> mark @MC_topOther;
                            pos ligature f_l
                                    <anchor NULL>
                                ligComponent
                                    <anchor 602 502> mark @MC_topOther;
                        } mark2liga_1;

                    } mark;
                    """
                ),
            ),
        ],
    )
    def test_multiple_anchor_classes_liga(self, FontClass, groupMarkClasses, expected):
        ufo = FontClass()
        liga = ufo.newGlyph("f_i")
        liga.appendAnchor({"name": "top_1", "x": 100, "y": 500})
        liga.appendAnchor({"name": "top_2", "x": 600, "y": 500})
        ligaOther = ufo.newGlyph("f_f")
        ligaOther.appendAnchor({"name": "topOther_1", "x": 101, "y": 501})
        ligaOther.appendAnchor({"name": "topOther_2", "x": 601, "y": 501})
        ligaMix = ufo.newGlyph("f_l")
        ligaMix.appendAnchor({"name": "top_1", "x": 102, "y": 502})
        ligaMix.appendAnchor({"name": "topOther_2", "x": 602, "y": 502})
        acutecomb = ufo.newGlyph("acutecomb")
        acutecomb.appendAnchor({"name": "_top", "x": 100, "y": 200})
        acutecomb.appendAnchor({"name": "_topOther", "x": 150, "y": 250})

        generated = self.writeFeatures(ufo, groupMarkClasses=groupMarkClasses)

        assert str(generated) == expected

    @pytest.mark.parametrize(
        "groupMarkClasses, warning, expected",
        [
            (
                True,
                # 'MC_top' should be last thanks to the anchorSortKey
                "MC_topOther, MC_top",
                dedent(
                    """\
                    markClass acutecomb <anchor 100 200> @MC_top;
                    markClass acutecomb <anchor 150 250> @MC_topOther;

                    feature mark {
                        lookup mark2base {
                            pos base a
                                <anchor 150 550> mark @MC_topOther;
                        } mark2base;

                        lookup mark2base_1 {
                            pos base a
                                <anchor 100 500> mark @MC_top;
                        } mark2base_1;

                    } mark;
                    """
                ),
            ),
            (
                False,
                # 'MC_topOther' (sorted alphabetically) should be last
                "MC_top, MC_topOther",
                dedent(
                    """\
                    markClass acutecomb <anchor 100 200> @MC_top;
                    markClass acutecomb <anchor 150 250> @MC_topOther;

                    feature mark {
                        lookup mark2base {
                            pos base a
                                <anchor 100 500> mark @MC_top;
                        } mark2base;

                        lookup mark2base_1 {
                            pos base a
                                <anchor 150 550> mark @MC_topOther;
                        } mark2base_1;

                    } mark;
                    """
                ),
            ),
        ],
    )
    def test_multiple_anchor_classes_conflict_warning(
        self, FontClass, caplog, groupMarkClasses, warning, expected
    ):
        """Check that when there is an ambiguity in the form of one base glyph
        and one mark glyph being able to be linked through two different
        anchor pairs, the mark feature writer emits a warning about the
        situation but still outputs a valid feature declaraction. The last
        lookup in that feature declaration will "win" and determine the outcome
        of mark positioning. See this comment for more information:
        https://github.com/googlefonts/ufo2ft/pull/416#issuecomment-721693266
        """
        caplog.set_level(logging.INFO)

        ufo = FontClass()
        liga = ufo.newGlyph("a")
        liga.appendAnchor({"name": "top", "x": 100, "y": 500})
        liga.appendAnchor({"name": "topOther", "x": 150, "y": 550})
        acutecomb = ufo.newGlyph("acutecomb")
        acutecomb.appendAnchor({"name": "_top", "x": 100, "y": 200})
        acutecomb.appendAnchor({"name": "_topOther", "x": 150, "y": 250})

        generated = self.writeFeatures(ufo, groupMarkClasses=groupMarkClasses)

        assert str(generated) == expected

    def test_skipExportGlyphs(self, testufo):
        testufo.lib["public.skipExportGlyphs"] = ["f_i", "tildecomb"]
        testufo.glyphOrder = ["a", "f_i", "acutecomb", "tildcomb"]

        generated = self.writeFeatures(testufo)

        assert str(generated) == dedent(
            """\
            markClass acutecomb <anchor 100 200> @MC_top;

            feature mark {
                lookup mark2base {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark2base;

            } mark;
            """
        )

    def test_quantize(self, testufo):
        testufo.newGlyph("ogonekcomb").anchors = [
            {"name": "_top", "x": 236, "y": 188},
        ]
        testufo.lib["public.skipExportGlyphs"] = ["f_i", "tildecomb"]
        generated = self.writeFeatures(testufo, quantization=50)

        assert str(generated) == dedent(
            """\
            markClass acutecomb <anchor 100 200> @MC_top;
            markClass ogonekcomb <anchor 250 200> @MC_top;

            feature mark {
                lookup mark2base {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark2base;

            } mark;
            """
        )

    def test_extra_substitutions(self, FontClass):
        dirname = os.path.dirname(os.path.dirname(__file__))
        fontPath = os.path.join(dirname, "data", "Alternates-Regular.ufo")
        testufo = FontClass(fontPath)
        generated = self.writeFeatures(
            testufo,
            compiler=FeatureCompiler(
                testufo,
                extraSubstitutions={
                    "uuMatra-oriya": {"uuMatra-oriya.BRACKET.varAlt01"},
                    "ka-oriya": {"ka-oriya.BRACKET.varAlt01"},
                    "lVocalicMatra-oriya": {"lVocalicMatra-oriya.BRACKET.varAlt01"},
                },
            ),
        )
        assert (
            "@MFS_blwm_mark2mark_bottom = [uuMatra-oriya lVocalicMatra-oriya "
            "ka-oriya.below lVocalicMatra-oriya.BRACKET.varAlt01 "
            "uuMatra-oriya.BRACKET.varAlt01]" in str(generated)
        )

    def test_contextual_anchors(self, FontClass):
        dirname = os.path.dirname(os.path.dirname(__file__))
        fontPath = os.path.join(dirname, "data", "ContextualAnchorsTest-Regular.ufo")
        testufo = FontClass(fontPath)

        writer = MarkFeatureWriter()
        feaFile = ast.FeatureFile()
        assert str(feaFile) == ""
        assert writer.write(testufo, feaFile)

        assert len(feaFile.markClasses) == 2
        assert "MC_bottom" in feaFile.markClasses

        feature = feaFile.statements[-1]
        assert feature.name == "mark"
        # note there are two mark2base lookups because ufo2ft v3 generates one lookup
        # per mark class (previously 'top' and 'bottom' would go into one lookup)
        assert str(feature) == dedent(
            """\
            feature mark {
                lookup mark2base;
                lookup mark2base_1;
                lookup ContextualMarkDispatch_0;
                lookup ContextualMarkDispatch_1;
                lookup ContextualMarkDispatch_2;
            } mark;
            """
        )

        lookup = feature.statements[-3].lookup
        assert str(lookup) == (
            "lookup ContextualMarkDispatch_0 {\n"
            "    lookupflag UseMarkFilteringSet [twodotshorizontalbelow-ar];\n"
            "    # reh-ar * behDotless-ar.medi &\n"
            "    pos reh-ar [behDotless-ar.init] behDotless-ar.medi"
            " @MC_bottom'"
            " lookup ContextualMark_0;\n"
            "} ContextualMarkDispatch_0;\n"
        )

        lookup = feature.statements[-2].lookup
        assert str(lookup) == (
            "lookup ContextualMarkDispatch_1 {\n"
            "    lookupflag UseMarkFilteringSet [twodotsverticalbelow-ar];\n"
            "    # dotbelow-ar *\n"
            "    pos dotbelow-ar [behDotless-ar.init.alt]"
            " @MC_bottom'"
            " lookup ContextualMark_1;\n"
            "    # reh-ar *\n"
            "    pos reh-ar [behDotless-ar.init behDotless-ar.init.alt]"
            " @MC_bottom'"
            " lookup ContextualMark_2;\n"
            "} ContextualMarkDispatch_1;\n"
        )

        lookup = feature.statements[-1].lookup
        assert str(lookup) == (
            "lookup ContextualMarkDispatch_2 {\n"
            "    # reh-ar *\n"
            "    pos reh-ar [behDotless-ar.init]"
            " @MC_bottom'"
            " lookup ContextualMark_3;\n"
            "} ContextualMarkDispatch_2;\n"
        )

    def test_contextual_anchors_no_mark_feature(self, testufo):
        a = testufo["a"]

        a.appendAnchor({"name": "*top", "x": 200, "y": 200, "identifier": "*top"})
        a.lib[OBJECT_LIBS_KEY] = {
            "*top": {
                "GPOS_Context": "f *",
            },
        }

        writer = MarkFeatureWriter(features=["mkmk"])
        feaFile = ast.FeatureFile()
        assert str(feaFile) == ""
        assert writer.write(testufo, feaFile)

        assert str(feaFile) == dedent(
            """\
            markClass acutecomb <anchor 100 200> @MC_top;
            markClass tildecomb <anchor 100 200> @MC_top;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;
            """
        )

        writer = MarkFeatureWriter()
        feaFile = ast.FeatureFile()
        assert str(feaFile) == ""
        assert writer.write(testufo, feaFile)

        assert str(feaFile) == dedent(
            """\
            markClass acutecomb <anchor 100 200> @MC_top;
            markClass tildecomb <anchor 100 200> @MC_top;

            lookup mark2base {
                pos base a
                    <anchor 100 200> mark @MC_top;
            } mark2base;

            lookup mark2liga {
                pos ligature f_i
                        <anchor 100 500> mark @MC_top
                    ligComponent
                        <anchor 600 500> mark @MC_top;
            } mark2liga;

            lookup ContextualMark_0 {
                pos base a
                    <anchor 200 200> mark @MC_top;
            } ContextualMark_0;

            lookup ContextualMarkDispatch_0 {
                # f *
                pos f [a] @MC_top' lookup ContextualMark_0;
            } ContextualMarkDispatch_0;

            feature mark {
                lookup mark2base;
                lookup mark2liga;
                lookup ContextualMarkDispatch_0;
            } mark;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;
            """
        )

    def test_contextual_liga_anchors(self, testufo):
        a = testufo["a"]

        a.appendAnchor({"name": "*top", "x": 200, "y": 200, "identifier": "*top"})
        a.lib[OBJECT_LIBS_KEY] = {
            "*top": {
                "GPOS_Context": "f *",
            },
        }

        fi = testufo["f_i"]
        fi.appendAnchor(
            {"name": "*top_1.tilde", "x": 300, "y": 500, "identifier": "*top_1.tilde"}
        )
        fi.appendAnchor(
            {"name": "*top_1.acute", "x": 200, "y": 300, "identifier": "*top_1.acute"}
        )
        fi.lib[OBJECT_LIBS_KEY] = {
            "*top_1.tilde": {
                "GPOS_Context": "* tildecomb",
            },
            "*top_1.acute": {
                "GPOS_Context": "* acutecomb",
            },
        }

        fl = testufo.newGlyph("f_l")
        fl.appendAnchor({"name": "top_1", "x": 200, "y": 400})
        fl.appendAnchor({"name": "top_2", "x": 500, "y": 400})
        fl.appendAnchor({"name": "*top_2", "x": 100, "y": 400, "identifier": "*top_2"})
        fl.lib[OBJECT_LIBS_KEY] = {
            "*top_2": {
                "GPOS_Context": "* tildecomb",
            },
        }

        writer = MarkFeatureWriter()
        feaFile = ast.FeatureFile()
        assert str(feaFile) == ""
        assert writer.write(testufo, feaFile)

        assert str(feaFile) == dedent(
            """\
            markClass acutecomb <anchor 100 200> @MC_top;
            markClass tildecomb <anchor 100 200> @MC_top;

            lookup mark2base {
                pos base a
                    <anchor 100 200> mark @MC_top;
            } mark2base;

            lookup mark2liga {
                pos ligature f_i
                        <anchor 100 500> mark @MC_top
                    ligComponent
                        <anchor 600 500> mark @MC_top;
                pos ligature f_l
                        <anchor 200 400> mark @MC_top
                    ligComponent
                        <anchor 500 400> mark @MC_top;
            } mark2liga;

            lookup ContextualMark_0 {
                pos base a
                    <anchor 200 200> mark @MC_top;
            } ContextualMark_0;

            lookup ContextualMark_1 {
                pos ligature f_i
                        <anchor 300 500> mark @MC_top
                    ligComponent
                        <anchor NULL>;
                pos ligature f_l
                        <anchor NULL>
                    ligComponent
                        <anchor 100 400> mark @MC_top;
            } ContextualMark_1;

            lookup ContextualMark_2 {
                pos ligature f_i
                        <anchor 200 300> mark @MC_top
                    ligComponent
                        <anchor NULL>;
            } ContextualMark_2;

            lookup ContextualMarkDispatch_0 {
                # f *
                pos f [a] @MC_top' lookup ContextualMark_0;
                # * tildecomb
                pos [f_i f_l] @MC_top' lookup ContextualMark_1 tildecomb;
                # * acutecomb
                pos [f_i] @MC_top' lookup ContextualMark_2 acutecomb;
            } ContextualMarkDispatch_0;

            feature mark {
                lookup mark2base;
                lookup mark2liga;
                lookup ContextualMarkDispatch_0;
            } mark;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;
            """
        )

    def test_contextual_liga_anchor_no_number(self, testufo):
        fi = testufo["f_i"]
        fi.appendAnchor(
            {"name": "*top.tilde", "x": 300, "y": 500, "identifier": "*top.tilde"}
        )
        fi.appendAnchor(
            {"name": "*top_1.acute", "x": 200, "y": 300, "identifier": "*top_1.acute"}
        )
        fi.lib[OBJECT_LIBS_KEY] = {
            "*top.tilde": {
                "GPOS_Context": "* tildecomb",
            },
            "*top_1.acute": {
                "GPOS_Context": "* acutecomb",
            },
        }

        writer = MarkFeatureWriter()
        feaFile = ast.FeatureFile()
        assert writer.write(testufo, feaFile)

        assert str(feaFile) == dedent(
            """\
            markClass acutecomb <anchor 100 200> @MC_top;
            markClass tildecomb <anchor 100 200> @MC_top;

            lookup mark2base {
                pos base a
                    <anchor 100 200> mark @MC_top;
            } mark2base;

            lookup mark2liga {
                pos ligature f_i
                        <anchor 100 500> mark @MC_top
                    ligComponent
                        <anchor 600 500> mark @MC_top;
            } mark2liga;

            lookup ContextualMark_0 {
                pos base f_i
                    <anchor 300 500> mark @MC_top;
            } ContextualMark_0;

            lookup ContextualMark_1 {
                pos ligature f_i
                        <anchor 200 300> mark @MC_top
                    ligComponent
                        <anchor NULL>;
            } ContextualMark_1;

            lookup ContextualMarkDispatch_0 {
                # * tildecomb
                pos [f_i] @MC_top' lookup ContextualMark_0 tildecomb;
                # * acutecomb
                pos [f_i] @MC_top' lookup ContextualMark_1 acutecomb;
            } ContextualMarkDispatch_0;

            feature mark {
                lookup mark2base;
                lookup mark2liga;
                lookup ContextualMarkDispatch_0;
            } mark;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;
            """
        )

    def test_contextual_mkmk_anchors(self, testufo):
        tildecomb = testufo["tildecomb"]

        tildecomb.appendAnchor(
            {"name": "*top", "x": 120, "y": 400, "identifier": "*top"}
        )
        tildecomb.lib[OBJECT_LIBS_KEY] = {
            "*top": {
                "GPOS_Context": "f *",
            },
        }

        writer = MarkFeatureWriter()
        feaFile = ast.FeatureFile()
        assert str(feaFile) == ""
        assert writer.write(testufo, feaFile)

        assert str(feaFile) == dedent(
            """\
            markClass acutecomb <anchor 100 200> @MC_top;
            markClass tildecomb <anchor 100 200> @MC_top;

            lookup mark2mark_top {
                @MFS_mark2mark_top = [acutecomb tildecomb];
                lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                pos mark tildecomb
                    <anchor 100 300> mark @MC_top;
            } mark2mark_top;

            lookup ContextualMarkToMark_0 {
                pos mark tildecomb
                    <anchor 120 400> mark @MC_top;
            } ContextualMarkToMark_0;

            lookup ContextualMarkToMarkDispatch_0 {
                # f *
                pos f [tildecomb] @MC_top' lookup ContextualMarkToMark_0;
            } ContextualMarkToMarkDispatch_0;

            feature mark {
                lookup mark2base {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark2base;

                lookup mark2liga {
                    pos ligature f_i
                            <anchor 100 500> mark @MC_top
                        ligComponent
                            <anchor 600 500> mark @MC_top;
                } mark2liga;

            } mark;

            feature mkmk {
                lookup mark2mark_top;
                lookup ContextualMarkToMarkDispatch_0;
            } mkmk;
            """
        )

    def test_contextual_anchor_no_context(self, testufo, caplog):
        a = testufo["a"]
        a.appendAnchor({"name": "*top", "x": 200, "y": 200, "identifier": "*top"})
        a.lib[OBJECT_LIBS_KEY] = {"*top": {ANCHOR_LIB_GPOS_CONTEXT_KEY: " "}}

        writer = MarkFeatureWriter()
        feaFile = ast.FeatureFile()
        assert str(feaFile) == ""

        logger = "ufo2ft.featureWriters.markFeatureWriter.MarkFeatureWriter"
        with caplog.at_level(logging.WARNING, logger=logger):
            assert writer.write(testufo, feaFile)
        assert len(caplog.records) == 1
        assert (
            "contextual anchor '*top' in glyph 'a' has no context data; skipped"
            in caplog.text
        )
        assert str(feaFile) == dedent(
            """\
            markClass acutecomb <anchor 100 200> @MC_top;
            markClass tildecomb <anchor 100 200> @MC_top;

            feature mark {
                lookup mark2base {
                    pos base a
                        <anchor 100 200> mark @MC_top;
                } mark2base;

                lookup mark2liga {
                    pos ligature f_i
                            <anchor 100 500> mark @MC_top
                        ligComponent
                            <anchor 600 500> mark @MC_top;
                } mark2liga;

            } mark;

            feature mkmk {
                lookup mark2mark_top {
                    @MFS_mark2mark_top = [acutecomb tildecomb];
                    lookupflag UseMarkFilteringSet @MFS_mark2mark_top;
                    pos mark tildecomb
                        <anchor 100 300> mark @MC_top;
                } mark2mark_top;

            } mkmk;
            """
        )

    def test_ignorable_anchors(self, FontClass):
        dirname = os.path.dirname(os.path.dirname(__file__))
        fontPath = os.path.join(dirname, "data", "IgnoreAnchorsTest-Thin.ufo")
        testufo = FontClass(fontPath)

        writer = MarkFeatureWriter()
        feaFile = ast.FeatureFile()
        assert str(feaFile) == ""
        assert writer.write(testufo, feaFile)

        assert len(feaFile.markClasses) == 1
        assert "MC_top" in feaFile.markClasses

        feature = feaFile.statements[-2]
        assert feature.name == "mark"
        assert len(feature.statements) == 1

        lookup = feature.statements[0]
        assert len(lookup.statements) == 4
        for statement in lookup.statements:
            assert isinstance(statement, ast.MarkBasePosStatement)
            assert len(statement.marks) == 1
            assert statement.marks[0][1].name == "MC_top"


if __name__ == "__main__":
    import sys

    sys.exit(pytest.main(sys.argv))



from dataclasses import dataclass
from typing import Optional, Type

from fontTools import varLib

from ufo2ft.constants import SPARSE_TTF_MASTER_TABLES
from ufo2ft.outlineCompiler import OutlineTTFCompiler
from ufo2ft.preProcessor import TTFInterpolatablePreProcessor
from ufo2ft.util import prune_unknown_kwargs

from .baseCompiler import BaseInterpolatableCompiler


@dataclass
class InterpolatableTTFCompiler(BaseInterpolatableCompiler):
    preProcessorClass: Type = TTFInterpolatablePreProcessor
    outlineCompilerClass: Type = OutlineTTFCompiler
    convertCubics: bool = True
    cubicConversionError: Optional[float] = None
    reverseDirection: bool = True
    flattenComponents: bool = False
    layerNames: Optional[str] = None
    colrLayerReuse: bool = False
    colrAutoClipBoxes: bool = False
    autoUseMyMetrics: bool = False
    allQuadratic: bool = True
    skipFeatureCompilation: bool = False

    def compileOutlines(self, ufo, glyphSet, layerName=None):
        kwargs = prune_unknown_kwargs(self.__dict__, self.outlineCompilerClass)
        kwargs["glyphDataFormat"] = 0 if self.allQuadratic else 1
        kwargs["tables"] = SPARSE_TTF_MASTER_TABLES if layerName else None
        # we want to keep coordinates as floats in glyf masters so that fonttools
        # can compute impliable on-curve points from unrounded coordinates before
        # building the VF
        kwargs["roundCoordinates"] = False
        # keep impliable oncurve points in the interpolatable master TTFs, they will
        # be pruned at the end by varLib in the final VF.
        kwargs["dropImpliedOnCurves"] = False
        outlineCompiler = self.outlineCompilerClass(ufo, glyphSet=glyphSet, **kwargs)
        return outlineCompiler.compile()

    def _merge(self, designSpaceDoc, excludeVariationTables):
        return varLib.build_many(
            designSpaceDoc,
            exclude=excludeVariationTables,
            optimize=self.optimizeGvar,
            skip_vf=lambda vf_name: self.variableFontNames
            and vf_name not in self.variableFontNames,
            colr_layer_reuse=self.colrLayerReuse,
            drop_implied_oncurves=self.dropImpliedOnCurves,
        )



import pytest
from fontTools.misc.loggingTools import CapturingLogHandler

from ufo2ft.filters.flattenComponents import FlattenComponentsFilter, logger


@pytest.fixture(
    params=[
        {
            "glyphs": [
                {"name": "space", "width": 500},
                {
                    "name": "contourGlyph",
                    "width": 350,
                    "outline": [
                        ("moveTo", ((0, 0),)),
                        ("lineTo", ((300, 0),)),
                        ("lineTo", ((300, 300),)),
                        ("lineTo", ((0, 300),)),
                        ("closePath", ()),
                    ],
                },
                {
                    "name": "componentGlyph",
                    "width": 350,
                    "outline": [("addComponent", ("contourGlyph", (1, 0, 0, 1, 0, 0)))],
                },
                {
                    "name": "nestedComponentGlyph",
                    "width": 350,
                    "outline": [
                        ("addComponent", ("componentGlyph", (1, 0, 0, 1, 0, 0)))
                    ],
                },
                {
                    "name": "componentAndNestedComponentsGlyph",
                    "width": 700,
                    "outline": [
                        ("addComponent", ("contourGlyph", (1, 0, 0, 1, 0, 0))),
                        ("addComponent", ("componentGlyph", (1, 0, 0, 1, 350, 0))),
                        (
                            "addComponent",
                            ("nestedComponentGlyph", (1, 0, 0, 1, 700, 0)),
                        ),
                    ],
                },
                {
                    "name": "contourAndComponentGlyph",
                    "width": 600,
                    "outline": [
                        ("moveTo", ((400, 0),)),
                        ("lineTo", ((400, 100),)),
                        ("lineTo", ((500, 100),)),
                        ("lineTo", ((500, 0),)),
                        ("closePath", ()),
                        ("addComponent", ("contourGlyph", (1, 0, 0, 1, 0, 0))),
                    ],
                },
                {
                    "name": "nestedContourAndComponentGlyph",
                    "width": 600,
                    "outline": [
                        (
                            "addComponent",
                            ("contourAndComponentGlyph", (1, 0, 0, 1, 50, 0)),
                        ),
                    ],
                },
                {
                    "name": "nestedNestedContourAndComponentGlyph",
                    "width": 600,
                    "outline": [
                        (
                            "addComponent",
                            ("nestedContourAndComponentGlyph", (1, 0, 0, 1, 45, 0)),
                        ),
                    ],
                },
                {
                    "name": "scaledComponentGlyph",
                    "width": 600,
                    "outline": [
                        (
                            "addComponent",
                            ("contourGlyph", (0.5, 0, 0, 0.5, 50, 50)),
                        ),
                    ],
                },
                {
                    "name": "nestedScaledComponentGlyph",
                    "width": 600,
                    "outline": [
                        (
                            "addComponent",
                            ("scaledComponentGlyph", (1, 0, 0, 1, 40, 40)),
                        ),
                    ],
                },
                {
                    "name": "scaledNestedComponentGlyph",
                    "width": 600,
                    "outline": [
                        (
                            "addComponent",
                            ("scaledComponentGlyph", (1.2, 0, 0, 1.2, 40, 40)),
                        ),
                    ],
                },
            ]
        }
    ]
)
def font(request, FontClass):
    font = FontClass()
    for param in request.param["glyphs"]:
        glyph = font.newGlyph(param["name"])
        glyph.width = param.get("width", 0)
        pen = glyph.getPen()
        for operator, operands in param.get("outline", []):
            getattr(pen, operator)(*operands)
    return font


class FlattenComponentsFilterTest:
    def test_empty_glyph(self, font):
        philter = FlattenComponentsFilter(include={"space"})
        assert not philter(font)

    def test_contour_glyph(self, font):
        philter = FlattenComponentsFilter(include={"contourGlyph"})
        assert not philter(font)

    def test_component_glyph(self, font):
        philter = FlattenComponentsFilter(include={"componentGlyph"})
        assert not philter(font)

    def test_nested_components_glyph(self, font):
        philter = FlattenComponentsFilter(include={"nestedComponentGlyph"})
        modified = philter(font)
        assert modified == {"nestedComponentGlyph"}
        assert [
            (c.baseGlyph, c.transformation)
            for c in font["nestedComponentGlyph"].components
        ] == [("contourGlyph", (1, 0, 0, 1, 0, 0))]

    def test_nested_contour_and_component_glyph(self, font):
        philter = FlattenComponentsFilter(
            include={
                "nestedContourAndComponentGlyph",
                "nestedNestedContourAndComponentGlyph",
            }
        )
        modified = philter(font)
        assert modified == {"nestedNestedContourAndComponentGlyph"}
        assert [
            (c.baseGlyph, c.transformation)
            for c in font["nestedNestedContourAndComponentGlyph"].components
        ] == [("contourAndComponentGlyph", (1, 0, 0, 1, 95, 0))]

    def test_scaled_component_glyph(self, font):
        philter = FlattenComponentsFilter(
            include={
                "scaledComponentGlyph",
                "nestedScaledComponentGlyph",
                "scaledNestedComponentGlyph",
            }
        )
        modified = philter(font)
        assert modified == {
            "nestedScaledComponentGlyph",
            "scaledNestedComponentGlyph",
        }
        assert [
            (c.baseGlyph, c.transformation)
            for c in font["nestedScaledComponentGlyph"].components
        ] == [("contourGlyph", (0.5, 0, 0, 0.5, 90, 90))]
        assert [
            (c.baseGlyph, c.transformation)
            for c in font["scaledNestedComponentGlyph"].components
        ] == [("contourGlyph", (0.6, 0, 0, 0.6, 100, 100))]

    def test_whole_font(self, font):
        philter = FlattenComponentsFilter()
        modified = philter(font)
        assert modified == {
            "nestedComponentGlyph",
            "componentAndNestedComponentsGlyph",
            "nestedNestedContourAndComponentGlyph",
            "nestedScaledComponentGlyph",
            "scaledNestedComponentGlyph",
        }
        assert [
            (c.baseGlyph, c.transformation)
            for c in font["nestedComponentGlyph"].components
        ] == [("contourGlyph", (1, 0, 0, 1, 0, 0))]
        assert [
            (c.baseGlyph, c.transformation)
            for c in font["componentAndNestedComponentsGlyph"].components
        ] == [
            ("contourGlyph", (1, 0, 0, 1, 0, 0)),
            ("contourGlyph", (1, 0, 0, 1, 350, 0)),
            ("contourGlyph", (1, 0, 0, 1, 700, 0)),
        ]
        assert [
            (c.baseGlyph, c.transformation)
            for c in font["nestedContourAndComponentGlyph"].components
        ] == [
            ("contourAndComponentGlyph", (1, 0, 0, 1, 50, 0)),
        ]
        assert [
            (c.baseGlyph, c.transformation)
            for c in font["nestedNestedContourAndComponentGlyph"].components
        ] == [("contourAndComponentGlyph", (1, 0, 0, 1, 95, 0))]
        assert [
            (c.baseGlyph, c.transformation)
            for c in font["nestedScaledComponentGlyph"].components
        ] == [("contourGlyph", (0.5, 0, 0, 0.5, 90, 90))]
        assert [
            (c.baseGlyph, c.transformation)
            for c in font["scaledNestedComponentGlyph"].components
        ] == [("contourGlyph", (0.6, 0, 0, 0.6, 100, 100))]

    def test_logger(self, font):
        with CapturingLogHandler(logger, level="INFO") as captor:
            philter = FlattenComponentsFilter()
            _ = philter(font)
        captor.assertRegex("Flattened composite glyphs: 5")



import pytest

from ufo2ft.filters import loadFilters
from ufo2ft.filters.dottedCircle import DottedCircleFilter
from ufo2ft.util import _GlyphSet


def test_dotted_circle_filter(FontClass, datadir):
    ufo_path = datadir.join("DottedCircleTest.ufo")
    font = FontClass(ufo_path)
    assert "uni25CC" not in font
    philter = DottedCircleFilter()
    glyphset = _GlyphSet.from_layer(font)

    modified = philter(font, glyphset)

    assert "uni25CC" in modified

    dotted_circle = glyphset["uni25CC"]

    # check the Glyph's module is the same as the Font's (both ufoLib2 or defcon,
    # not mixed): https://github.com/googlefonts/ufo2ft/issues/644
    font_ufo_module = type(font).__module__.split(".")[0]
    glyph_ufo_module = type(dotted_circle).__module__.split(".")[0]
    assert glyph_ufo_module == font_ufo_module

    anchors = list(sorted(dotted_circle.anchors, key=lambda x: x.name))
    assert anchors[0].x == 464
    assert anchors[0].y == -17
    assert anchors[0].name == "bottom"

    assert anchors[1].x == 563
    assert anchors[1].y == 546
    assert anchors[1].name == "top"

    assert len(dotted_circle) == 12
    assert int(dotted_circle.width) == 688
    assert dotted_circle.unicodes == [0x25CC]


def test_empty_font(FontClass):
    """Check that the filter works on an empty font, i.e. uses fallbacks where
    appropriate."""

    font = FontClass()
    font.lib["com.github.googlei18n.ufo2ft.filters"] = [
        {"name": "dottedCircle", "pre": True}
    ]

    pre_filters, _ = loadFilters(font)
    (philter,) = pre_filters
    glyphset = _GlyphSet.from_layer(font)

    modified = philter(font, glyphset)

    assert "uni25CC" in modified


@pytest.mark.filterwarnings("ignore:Please update")
def test_empty_font_deprecated(FontClass):
    """Check that the module redirection works."""

    font = FontClass()
    font.lib["com.github.googlei18n.ufo2ft.filters"] = [
        {"name": "DottedCircleFilter", "pre": True}
    ]

    pre_filters, _ = loadFilters(font)
    (philter,) = pre_filters
    glyphset = _GlyphSet.from_layer(font)

    modified = philter(font, glyphset)

    assert "uni25CC" in modified



from fontTools.pens.pointPen import ReverseContourPointPen

from ufo2ft.filters import BaseFilter


class ReverseContourDirectionFilter(BaseFilter):
    def filter(self, glyph):
        if not len(glyph):
            return False
        pen = ReverseContourPointPen(glyph.getPointPen())
        contours = list(glyph)
        glyph.clearContours()
        for contour in contours:
            contour.drawPoints(pen)
        return True



import itertools
import re
from collections import OrderedDict, defaultdict
from functools import partial
from typing import Dict, Optional, Set, Tuple

from ufo2ft.constants import (
    ANCHOR_LIB_GPOS_CONTEXT_KEY,
    INDIC_SCRIPTS,
    OBJECT_LIBS_KEY,
    USE_SCRIPTS,
)
from ufo2ft.featureWriters import BaseFeatureWriter, ast
from ufo2ft.util import (
    classifyGlyphs,
    otRoundIgnoringVariable,
    unicodeInScripts,
    unicodeScriptExtensions,
)


class AbstractMarkPos:
    """Object containing all the mark attachments for glyph 'name'.
    The 'marks' is a list of NamedAnchor objects.
    Provides methods to filter marks given some callable, and convert
    itself to feaLib AST 'pos' statements for mark2base, mark2liga and
    mark2mark lookups.
    """

    Statement = None

    def __init__(self, name, marks):
        self.name = name
        self.marks = marks

    def _filterMarks(self, include):
        return [anchor for anchor in self.marks if include(anchor)]

    def _marksAsAST(self):
        return [
            (
                ast.Anchor(
                    x=otRoundIgnoringVariable(anchor.x),
                    y=otRoundIgnoringVariable(anchor.y),
                ),
                anchor.markClass,
            )
            for anchor in sorted(self.marks, key=lambda a: a.name)
        ]

    def asAST(self):
        marks = self._marksAsAST()
        return self.Statement(ast.GlyphName(self.name), marks)

    def __str__(self):
        return self.asAST().asFea()  # pragma: no cover

    def filter(self, include):
        marks = self._filterMarks(include)
        return self.__class__(self.name, marks) if any(marks) else None

    def getMarkGlyphToMarkClasses(self):
        """Return a list of pairs (markGlyph, markClasses)."""
        markGlyphToMarkClasses = defaultdict(set)
        for namedAnchor in self.marks:
            for markGlyph in namedAnchor.markClass.glyphs:
                markGlyphToMarkClasses[markGlyph].add(namedAnchor.markClass.name)
        return markGlyphToMarkClasses.items()


class MarkToBasePos(AbstractMarkPos):
    Statement = ast.MarkBasePosStatement


class MarkToMarkPos(AbstractMarkPos):
    Statement = ast.MarkMarkPosStatement


class MarkToLigaPos(AbstractMarkPos):
    Statement = ast.MarkLigPosStatement

    def _filterMarks(self, include):
        return [
            [anchor for anchor in component if include(anchor)]
            for component in self.marks
        ]

    def _marksAsAST(self):
        return [
            [
                (
                    ast.Anchor(
                        x=otRoundIgnoringVariable(anchor.x),
                        y=otRoundIgnoringVariable(anchor.y),
                    ),
                    anchor.markClass,
                )
                for anchor in sorted(component, key=lambda a: a.name)
            ]
            for component in self.marks
        ]

    def getMarkGlyphToMarkClasses(self):
        """Return a list of pairs (markGlyph, markClasses)."""
        markGlyphToMarkClasses = defaultdict(set)
        for component in self.marks:
            for namedAnchor in component:
                for markGlyph in namedAnchor.markClass.glyphs:
                    markGlyphToMarkClasses[markGlyph].add(namedAnchor.markClass.name)
        return markGlyphToMarkClasses.items()


MARK_PREFIX = LIGA_SEPARATOR = "_"
LIGA_NUM_RE = re.compile(r".*?(\d+)$")


def parseAnchorName(
    anchorName,
    markPrefix=MARK_PREFIX,
    ligaSeparator=LIGA_SEPARATOR,
    ligaNumRE=LIGA_NUM_RE,
    ignoreRE=None,
    libData=None,
):
    """Parse anchor name and return a tuple that specifies:
    1) whether the anchor is a "mark" anchor (bool);
    2) the "key" name of the anchor, i.e. the name after stripping all the
       prefixes and suffixes, which identifies the class it belongs to (str);
    3) An optional number (int), starting from 1, which identifies that index
       of the ligature component the anchor refers to.

    The 'ignoreRE' argument is an optional regex pattern (str) identifying
    sub-strings in the anchor name that should be ignored when parsing the
    three elements above.
    """
    number = None
    isContextual = False
    if ignoreRE is not None:
        anchorName = re.sub(ignoreRE, "", anchorName)

    if anchorName[0] == "*" and libData and ANCHOR_LIB_GPOS_CONTEXT_KEY in libData:
        isContextual = True
        anchorName = anchorName[1:]
        anchorName = re.sub(r"\..*", "", anchorName)

    m = ligaNumRE.match(anchorName)
    if not m:
        key = anchorName
    else:
        number = m.group(1)
        key = anchorName.rstrip(number)
        separator = ligaSeparator
        if key.endswith(separator):
            assert separator
            key = key[: -len(separator)]
            number = int(number)
        else:
            # not a valid ligature anchor name
            key = anchorName
            number = None

    if anchorName.startswith(markPrefix) and key:
        if number is not None:
            raise ValueError("mark anchor cannot be numbered: %r" % anchorName)
        isMark = True
        key = key[len(markPrefix) :]
        if not key:
            raise ValueError("mark anchor key is nil: %r" % anchorName)
    else:
        isMark = False

    isIgnorable = key and not key[0].isalpha()

    return isMark, key, number, isContextual, isIgnorable


class NamedAnchor:
    """A position with a name, and an associated markClass."""

    __slots__ = (
        "name",
        "x",
        "y",
        "isMark",
        "key",
        "number",
        "markClass",
        "isContextual",
        "isIgnorable",
        "libData",
    )

    # subclasses can customize these to use different anchor naming schemes
    markPrefix = MARK_PREFIX
    ignoreRE = None
    ligaSeparator = LIGA_SEPARATOR
    ligaNumRE = LIGA_NUM_RE

    def __init__(self, name, x, y, markClass=None, libData=None):
        self.name = name
        self.x = x
        self.y = y
        isMark, key, number, isContextual, isIgnorable = parseAnchorName(
            name,
            markPrefix=self.markPrefix,
            ligaSeparator=self.ligaSeparator,
            ligaNumRE=self.ligaNumRE,
            ignoreRE=self.ignoreRE,
            libData=libData,
        )
        if number is not None:
            if number < 1:
                raise ValueError("ligature component indexes must start from 1")
        else:
            assert key, name
        self.isMark = isMark
        self.key = key
        self.number = number
        self.markClass = markClass
        self.isContextual = isContextual
        self.isIgnorable = isIgnorable
        self.libData = libData

    @property
    def markAnchorName(self):
        return self.markPrefix + self.key

    def __repr__(self):
        items = ("{}={!r}".format(k, getattr(self, k)) for k in ("name", "x", "y"))
        return "%s(%s)" % (type(self).__name__, ", ".join(items))


def colorGraph(adjacency):
    """Color the graph defined by the provided adjacency lists.
    The input is a dict of iterables. Each entry of the dict is one vertex,
    and the value is a list of neighbours of that vertex.
    The input graph is expected to be undirected and the input should reflect
    that (have symmetric adjacency for A -> B and B -> A).
    Vertices that don't have neighbours should still be present in the input.

    The output is a list of lists, each list being one color assignment,
    and its members being vertices.
    """
    # Basic implementation
    # https://en.wikipedia.org/wiki/Greedy_coloring
    colors = dict()
    # Sorted for reproducibility, probably not the optimal vertex order
    for node in sorted(adjacency):
        usedNeighbourColors = {
            colors[neighbour] for neighbour in adjacency[node] if neighbour in colors
        }
        colors[node] = firstAvailable(usedNeighbourColors)
    groups = defaultdict(list)
    for node, color in colors.items():
        groups[color].append(node)
    return list(groups.values())


def firstAvailable(colorSet):
    """Return smallest non-negative integer not in the given set of colors."""
    count = 0
    while True:
        if count not in colorSet:
            return count
        count += 1


class MarkFeatureWriter(BaseFeatureWriter):
    """Generates a mark, mkmk, abvm and blwm features based on glyph anchors.

    The default mode is 'skip': i.e. if any of the supported features is
    already present in the feature file, it is not generated again.

    The optional 'append' mode will add extra lookups to already existing
    features, if any. New markClass definitions with unique names are
    generated when the mark anchors from UFO glyphs are different from those
    already defined in the feature file, otherwise the existing markClass
    definitions are reused in the newly appended lookups.

    Anchors prefixed with "_" are considered mark anchors; any glyph
    containing those is as such considered a mark glyph, thus added to
    markClass definitions, and in mark-to-mark lookups (if the glyph also
    contains other non-underscore-prefixed anchors).

    Anchors suffixed with a number, e.g. "top_1", "bottom_2", etc., are used
    for ligature glyphs. The number refers to the index (counting from 1) of
    the ligature component where the mark is meant to be attached.

    It is possible that a ligature component has no marks defined, in which
    case one can have an anchor with an empty name and only the number (e.g.
    '_3'), which is encoded as '<anchor NULL>' in the generated 'pos ligature'
    statement.

    If the glyph set contains glyphs whose unicode codepoint's script extension
    property intersects with one of the scripts which are processed by the Indic,
    USE, or Khmer complex shapers, then the "abvm" and "blwm" features are also
    generated for those glyphs, as well as for alternate glyphs only accessible
    via GSUB substitutions.

    The "abvm" (above-base marks) and "blwm" (below-base marks) features
    include all mark2base, mark2liga and mark2mark attachments for Indic/USE/Khmer
    glyphs containing anchors from predefined lists of "above" and "below" anchor
    names (see below). If these glyphs contain anchors with names not in those
    lists, the anchors' vertical position relative to the half of the UPEM
    square is used to decide whether they are considered above or below.

    If the `quantization` argument is given in the filter options, the resulting
    anchors are rounded to the nearest multiple of the quantization value.

    If `groupMarkClases=True`, mark-to-base or mark-to-ligature attachments that
    reference non-overlapping mark classes will get grouped in the same lookup; and
    if a mark glyph is in more than one mark class, additional lookups will be generated
    for those as required. NOTE: this was the default behavior until ufo2ft 2.33.4.
    The current default behavior was simplified to match other font editors and
    we now build as many mark-to-base and mark-to-liga lookups as there
    are mark classes, and lookups are sorted alphabetically by the mark class
    name so the more specific ('top.alt' instead 'top') would be applied last and
    wins in case when the same base or ligature glyph can attach to the same mark
    through multiple mark classes.
    https://github.com/googlefonts/ufo2ft/issues/591
    """

    options = dict(quantization=1, groupMarkClasses=False)

    tableTag = "GPOS"
    features = frozenset(["mark", "mkmk", "abvm", "blwm"])

    # subclasses may override this to use different anchor naming schemes
    NamedAnchor = NamedAnchor

    # @MC_top, @MC_bottom, etc.
    markClassPrefix = "MC"

    abvmAnchorNames = {
        "top",
        "topleft",
        "topright",
        "candra",
        "bindu",
        "candrabindu",
        "imatra",
    }
    blwmAnchorNames = {"bottom", "bottomleft", "bottomright", "nukta"}

    scriptsUsingAbvm = set(INDIC_SCRIPTS + USE_SCRIPTS + ["Khmr"])

    # Glyphs moves "_bottom" and "_top" (if present) to the top of
    # the list and then picks the first to use in the mark feature.
    # https://github.com/googlei18n/noto-source/issues/122#issuecomment-403952188
    anchorSortKey = {"_bottom": -2, "_top": -1}

    def setContext(self, font, feaFile, compiler=None):
        ctx = super().setContext(font, feaFile, compiler=compiler)
        ctx.gdefClasses = self.getGDEFGlyphClasses()
        ctx.anchorLists = self._getAnchorLists()
        ctx.anchorPairs = self._getAnchorPairs()
        ctx.feaScripts = set(ast.getScriptLanguageSystems(feaFile).keys())

    def shouldContinue(self):
        if not self.context.anchorPairs:
            self.log.debug("No mark-attaching anchors found; skipped")
            return False
        return super().shouldContinue()

    def _getAnchorLists(self):
        gdefClasses = self.context.gdefClasses
        if gdefClasses.base is not None:
            # only include the glyphs listed in the GDEF.GlyphClassDef groups
            include = gdefClasses.base | gdefClasses.ligature | gdefClasses.mark
        else:
            # no GDEF table defined in feature file, include all glyphs
            include = None
        result = OrderedDict()
        for glyphName, glyph in self.getOrderedGlyphSet().items():
            if include is not None and glyphName not in include:
                continue
            anchorDict = OrderedDict()
            objectLibs = glyph.lib.get(OBJECT_LIBS_KEY)
            for anchor in glyph.anchors:
                anchorName = anchor.name
                if not anchorName:
                    self.log.warning(
                        "unnamed anchor discarded in glyph '%s'", glyphName
                    )
                    continue
                if anchorName in anchorDict:
                    self.log.warning(
                        "duplicate anchor '%s' in glyph '%s'", anchorName, glyphName
                    )
                x, y = self._getAnchor(glyphName, anchorName, anchor=anchor)
                libData = None
                if anchor.identifier and objectLibs:
                    libData = objectLibs.get(anchor.identifier)
                a = self.NamedAnchor(name=anchorName, x=x, y=y, libData=libData)
                if a.isIgnorable:
                    continue
                anchorDict[anchorName] = a
            if anchorDict:
                result[glyphName] = list(anchorDict.values())
        return result

    def _getAnchorPairs(self):
        markAnchorNames = set()
        for anchors in self.context.anchorLists.values():
            markAnchorNames.update(a.name for a in anchors if a.isMark)
        anchorPairs = {}
        for anchors in self.context.anchorLists.values():
            for anchor in anchors:
                if anchor.isMark:
                    continue
                markAnchorName = anchor.markAnchorName
                if markAnchorName in markAnchorNames:
                    anchorPairs[anchor.name] = markAnchorName
        return anchorPairs

    def _pruneUnusedAnchors(self):
        baseAnchorNames = set(self.context.anchorPairs.keys())
        markAnchorNames = set(self.context.anchorPairs.values())
        attachingAnchorNames = baseAnchorNames | markAnchorNames
        for glyphName, anchors in list(self.context.anchorLists.items()):
            for anchor in list(anchors):
                if anchor.name not in attachingAnchorNames and anchor.key:
                    anchors.remove(anchor)
            if not anchors:
                del self.context.anchorLists[glyphName]

    def _groupMarkGlyphsByAnchor(self):
        gdefMarks = self.context.gdefClasses.mark
        markAnchorNames = set(self.context.anchorPairs.values())
        markGlyphNames = set()
        groups = {}
        for glyphName, anchors in self.context.anchorLists.items():
            # if the feature file has a GDEF table with GlyphClassDef defined,
            # only include mark glyphs that are referenced in there, otherwise
            # include any glyphs that contain an "_" prefixed anchor.
            if gdefMarks is not None and glyphName not in gdefMarks:
                continue
            markAnchors = [a for a in anchors if a.name in markAnchorNames]
            if not markAnchors:
                continue
            # Use all mark anchors. The rest of the algorithm will make sure
            # that the generated lookups will not have overlapping mark classes.
            for anchor in markAnchors:
                group = groups.setdefault(anchor.name, OrderedDict())
                assert glyphName not in group
                group[glyphName] = anchor
            markGlyphNames.add(glyphName)
        self.context.markGlyphNames = markGlyphNames
        return groups

    def _makeMarkClassDefinitions(self):
        markGlyphSets = self._groupMarkGlyphsByAnchor()
        currentClasses = self.context.feaFile.markClasses
        allMarkClasses = self.context.markClasses = {}
        classPrefix = self.markClassPrefix
        newDefs = []
        for markAnchorName, glyphAnchorPairs in sorted(markGlyphSets.items()):
            className = ast.makeFeaClassName(classPrefix + markAnchorName)
            for glyphName, anchor in glyphAnchorPairs.items():
                mcd = self._defineMarkClass(
                    glyphName, anchor.x, anchor.y, className, currentClasses
                )
                if mcd is not None:
                    newDefs.append(mcd)
                    # this may be different because of name clashes
                    className = mcd.markClass.name
                allMarkClasses[anchor.key] = currentClasses[className]
        return newDefs

    def _defineMarkClass(self, glyphName, x, y, className, markClasses):
        anchor = ast.Anchor(x=otRoundIgnoringVariable(x), y=otRoundIgnoringVariable(y))
        markClass = markClasses.get(className)
        if markClass is None:
            markClass = ast.MarkClass(className)
            markClasses[className] = markClass
        else:
            if glyphName in markClass.glyphs:
                mcdef = markClass.glyphs[glyphName]
                if self._anchorsAreEqual(anchor, mcdef.anchor):
                    self.log.debug(
                        "Glyph %s already defined in markClass @%s",
                        glyphName,
                        className,
                    )
                    return None
                else:
                    # same mark glyph defined with different anchors for the
                    # same markClass; make a new unique markClass definition
                    newClassName = ast.makeFeaClassName(className, markClasses)
                    markClass = ast.MarkClass(newClassName)
                    markClasses[newClassName] = markClass
        glyphName = ast.GlyphName(glyphName)
        mcdef = ast.MarkClassDefinition(markClass, anchor, glyphName)
        markClass.addDefinition(mcdef)
        return mcdef

    @staticmethod
    def _anchorsAreEqual(a1, a2):
        # TODO add __eq__ to feaLib AST objects?
        return all(
            getattr(a1, attr) == getattr(a2, attr)
            for attr in ("x", "y", "contourpoint", "xDeviceTable", "yDeviceTable")
        )

    def _setBaseAnchorMarkClasses(self):
        markClasses = self.context.markClasses
        for anchors in self.context.anchorLists.values():
            for anchor in anchors:
                if anchor.isMark or not anchor.key or anchor.key not in markClasses:
                    continue
                anchor.markClass = markClasses[anchor.key]

    def _groupMarkClasses(self, markGlyphToMarkClasses):
        # To compute the number of lookups that we need to build, we want
        # the minimum number of lookups such that, whenever a mark glyph
        # belongs to several mark classes, these classes are not in the same
        # lookup. A trivial solution is to make 1 lookup per mark class
        # but that's a bit wasteful, we might be able to do better by grouping
        # mark classes that do not conflict.
        # This is a graph coloring problem: the graph nodes are mark classes,
        # edges are between classes that would conflict and the colors are
        # the lookups in which they can go.
        adjacency = {
            # We'll get the same markClass several times in the dict
            # comprehension below but it's ok, only one will be kept.
            markClass: set()
            for markClasses in markGlyphToMarkClasses.values()
            for markClass in markClasses
        }
        for _markGlyph, markClasses in markGlyphToMarkClasses.items():
            for markClass, other in itertools.combinations(markClasses, 2):
                adjacency[markClass].add(other)
                adjacency[other].add(markClass)
        colorGroups = colorGraph(adjacency)
        # Sort the groups, because the group that contains MC_top or MC_bottom
        # needs to go to the end (as specified in self.anchorSortKey) so that
        # they are applied last and "win" in case of conflict.
        # We also sort alphabetically for reproducibility, both within each
        # group and between groups.
        return sorted(
            [sorted(group) for group in colorGroups],
            key=lambda group: (
                # The first part sorts _top and _bottom at the end.
                # There's a minus sign in front of the min because the original
                # self.anchorSortKey was designed to put the _top and _bottom
                # at the start (and now we want them at the end).
                -min(
                    # Remove the MC prefix because that's how the mark classes
                    # are looking at this stage (the original
                    # self.anchorSortKey was applied at a different stage of
                    # the algorithm, on anchors instead of mark classes)
                    self.anchorSortKey.get(self._removeClassPrefix(markClass), 0)
                    for markClass in group
                ),
                # Second part of the tuple sorts the groups lexicographically
                group,
            ),
        )

    def _removeClassPrefix(self, markClass):
        assert markClass.startswith(self.markClassPrefix)
        return markClass[len(self.markClassPrefix) :]

    def _groupAttachments(self, attachments):
        """Group the given attachments so that no group contains conflicting
        anchor classes for the same glyph.
        """
        # Idea for mark2base:
        #   attachments is a list of mark to base pairs, linked together through
        #   an anchor name We have to put them into one or more lookups with the
        #   constraint that the same mark glyph cannot appear twice in the same
        #   lookup while using different anchor names.
        # Idea for mark2liga:
        #   attachments is a list of mark to liga positioning. Each links
        #   together a base ligature with several marks, through numbered anchor
        #   names.
        #   We have to put them into one or more lookups with the constraint that
        #   the same mark glyph cannot appear twice in the same lookup while
        #   using different anchor names.
        #   To do so, if a single attachment refers to to the same mark twice
        #   through different anchor names, we may have to split the attachment
        #   into two attachments, using null anchors instead of one or the other
        #   mark class in each split attachment.
        if self.options.groupMarkClasses:
            markGlyphToMarkClasses = defaultdict(set)
            for attachment in attachments:
                for markGlyph, markClasses in attachment.getMarkGlyphToMarkClasses():
                    markGlyphToMarkClasses[markGlyph].update(markClasses)
            groupedMarkClasses = self._groupMarkClasses(markGlyphToMarkClasses)
        else:
            # this will generate one lookup per mark class, and sort them
            # lexicographically by the anchor name, so the lookup for e.g.
            # '_top.alt01' will occur *after* the one for `_top` (the last wins) thus
            # allowing some degree of control on potentially ambiguous attachments
            # https://github.com/googlefonts/ufo2ft/issues/762
            # https://github.com/googlefonts/ufo2ft/issues/591
            groupedMarkClasses = [
                [markClass.name]
                for _, markClass in sorted(self.context.markClasses.items())
            ]
        lookups = []
        for markClasses in groupedMarkClasses:
            lookup = []
            # Filter existing attachments
            for attachment in attachments:
                # One attachment has one base glyph and many marks, each of
                # the class NamedAnchor. Each NamedAnchor has one markClass.
                # We keep the NamedAnchor if the markClass is allowed in the
                # current lookup.
                def include(anchor):
                    return anchor.markClass.name in markClasses  # noqa: B023

                filteredAttachment = attachment.filter(include)
                if filteredAttachment:
                    lookup.append(filteredAttachment)
            lookups.append(lookup)
        return lookups

    def _makeMarkToBaseAttachments(self):
        markGlyphNames = self.context.markGlyphNames
        baseClass = self.context.gdefClasses.base
        result = []
        for glyphName, anchors in self.context.anchorLists.items():
            # exclude mark glyphs, or glyphs not listed in GDEF Base
            if glyphName in markGlyphNames or (
                baseClass is not None and glyphName not in baseClass
            ):
                continue
            baseMarks = []
            for anchor in anchors:
                if anchor.markClass is None or anchor.number is not None:
                    # skip anchors for which no mark class is defined; also
                    # skip '_1', '_2', etc. suffixed anchors for this lookup
                    # type; these will be are added in the mark2liga lookup
                    continue
                if anchor.isContextual:
                    # skip contextual anchors. They are handled separately.
                    continue
                assert not anchor.isMark
                baseMarks.append(anchor)
            if not baseMarks:
                continue
            result.append(MarkToBasePos(glyphName, baseMarks))
        return result

    def _makeMarkToMarkAttachments(self):
        markGlyphNames = self.context.markGlyphNames
        # we make a dict of lists containing mkmk pos rules keyed by
        # anchor name, so we can create one mkmk lookup per markClass
        # each with different mark filtering sets.
        results = {}
        for glyphName, anchors in self.context.anchorLists.items():
            if glyphName not in markGlyphNames:
                continue
            for anchor in anchors:
                # skip anchors for which no mark class is defined
                if anchor.markClass is None or anchor.isMark:
                    continue
                if anchor.isContextual:
                    # skip contextual anchors. They are handled separately.
                    continue
                if anchor.number is not None:
                    self.log.warning(
                        "invalid ligature anchor '%s' in mark glyph '%s'; " "skipped",
                        anchor.name,
                        glyphName,
                    )
                    continue
                pos = MarkToMarkPos(glyphName, [anchor])
                results.setdefault(anchor.key, []).append(pos)
        return results

    def _makeMarkToLigaAttachments(self):
        markGlyphNames = self.context.markGlyphNames
        ligatureClass = self.context.gdefClasses.ligature
        result = []
        for glyphName, anchors in self.context.anchorLists.items():
            # exclude mark glyphs, or glyphs not listed in GDEF Ligature
            if glyphName in markGlyphNames or (
                ligatureClass is not None and glyphName not in ligatureClass
            ):
                continue
            componentAnchors = {}
            for anchor in anchors:
                if anchor.markClass is None and anchor.key:
                    # skip anchors for which no mark class is defined
                    continue
                assert not anchor.isMark
                number = anchor.number
                if number is None:
                    # we handled these in the mark2base lookup
                    continue
                if anchor.isContextual:
                    # skip contextual anchors. They are handled separately.
                    continue
                # unnamed anchors with only a number suffix "_1", "_2", etc.
                # are understood as the ligature component having <anchor NULL>
                if not anchor.key:
                    componentAnchors[number] = []
                else:
                    componentAnchors.setdefault(number, []).append(anchor)
            if not componentAnchors:
                continue
            ligatureMarks = []
            # ligature components are indexed from 1; any missing intermediate
            # anchor number means the component has <anchor NULL>
            for number in range(1, max(componentAnchors.keys()) + 1):
                ligatureMarks.append(componentAnchors.get(number, []))
            result.append(MarkToLigaPos(glyphName, ligatureMarks))
        return result

    def _makeContextualAttachments(
        self,
        baseClass: Optional[Set[str]],
        ligatureClass: Optional[Set[str]],
        markClass: Optional[Set[str]],
    ) -> Tuple[Dict[str, Tuple[str, NamedAnchor]], Dict[str, Tuple[str, NamedAnchor]]]:
        def includedOrNoClass(gdefClass: Optional[Set[str]], glyphName: str) -> bool:
            return glyphName in gdefClass if gdefClass is not None else True

        def includedInClass(gdefClass: Optional[Set[str]], glyphName: str) -> bool:
            return glyphName in gdefClass if gdefClass is not None else False

        abvmGlyphs, _ = self._getAbvmGlyphs()
        markGlyphNames = self.context.markGlyphNames

        baseResult = defaultdict(list)
        ligatureResult = defaultdict(list)
        markResult = defaultdict(list)
        abvmResult = defaultdict(list)

        for glyphName, anchors in sorted(self.context.anchorLists.items()):
            for anchor in anchors:
                # Skip non-contextual anchors
                if not anchor.isContextual:
                    continue

                # Mark glyphs go to mkmk lookups
                if glyphName in markGlyphNames:
                    # skip anchors for which no mark class is defined
                    if anchor.markClass is None or anchor.isMark:
                        continue
                    if anchor.number is not None:
                        self.log.warning(
                            "invalid contextual ligature anchor '%s' in mark glyph '%s'; "
                            "skipped",
                            anchor.name,
                            glyphName,
                        )
                        continue
                    dest = markResult
                # See "after" truth table for what this logic hopes to achieve:
                # https://github.com/googlefonts/ufo2ft/pull/890#issuecomment-2498032081
                elif anchor.number is not None and includedOrNoClass(
                    ligatureClass, glyphName
                ):
                    dest = ligatureResult
                elif anchor.number is None and (
                    includedOrNoClass(baseClass, glyphName)
                    or includedInClass(ligatureClass, glyphName)
                ):
                    dest = abvmResult if glyphName in abvmGlyphs else baseResult
                else:
                    continue

                anchor_context = anchor.libData[ANCHOR_LIB_GPOS_CONTEXT_KEY].strip()

                if not anchor_context:
                    self.log.warning(
                        "contextual anchor '%s' in glyph '%s' has no context data; skipped",
                        anchor.name,
                        glyphName,
                    )
                    continue

                for context in anchor_context.splitlines():
                    context = context.strip().rstrip(";")
                    if not context:
                        continue
                    dest[context].append((glyphName, anchor))
        return baseResult, ligatureResult, markResult, abvmResult

    @staticmethod
    def _iterAttachments(attachments, include=None, marksFilter=None):
        for pos in attachments:
            if include is not None and not include(pos.name):
                continue
            if marksFilter is not None:
                pos = pos.filter(marksFilter)
                if pos is None:
                    continue
            yield pos

    def _makeMarkLookup(self, lookupName, attachments, include, marksFilter=None):
        statements = [
            pos.asAST()
            for pos in self._iterAttachments(attachments, include, marksFilter)
        ]
        if statements:
            lkp = ast.LookupBlock(lookupName)
            lkp.statements.extend(statements)
            return lkp

    def _makeMarkFilteringSetClass(self, lookupName, attachments, markClass, include):
        markGlyphs = (glyphName for glyphName in markClass.glyphs if include(glyphName))
        baseGlyphs = (
            pos.name for pos in attachments if pos.name not in markClass.glyphs
        )
        members = itertools.chain(markGlyphs, baseGlyphs)
        className = "MFS_%s" % lookupName
        return ast.makeGlyphClassDefinitions(
            {className: members}, feaFile=self.context.feaFile
        )[className]

    def _makeMarkToMarkLookup(
        self, anchorName, attachments, include, marksFilter=None, featureTag=None
    ):
        attachments = list(self._iterAttachments(attachments, include, marksFilter))
        if not attachments:
            return
        prefix = (featureTag + "_") if featureTag is not None else ""
        lookupName = f"{prefix}mark2mark_{anchorName}"
        filteringClass = self._makeMarkFilteringSetClass(
            lookupName,
            attachments,
            markClass=self.context.markClasses[anchorName],
            include=include,
        )
        lkp = ast.LookupBlock(lookupName)
        lkp.statements.append(filteringClass)
        lkp.statements.append(ast.makeLookupFlag(markFilteringSet=filteringClass))
        lkp.statements.extend(pos.asAST() for pos in attachments)
        return lkp

    def _makeMarkFeature(self, include):
        # First make the non-contextual lookups
        baseLkps = []
        for attachments in self.context.groupedMarkToBaseAttachments:
            i = len(baseLkps)
            lookup = self._makeMarkLookup(
                f"mark2base{'_' + str(i) if i > 0 else ''}", attachments, include
            )
            if lookup:
                baseLkps.append(lookup)
        ligaLkps = []
        for attachments in self.context.groupedMarkToLigaAttachments:
            i = len(ligaLkps)
            lookup = self._makeMarkLookup(
                f"mark2liga{'_' + str(i) if i > 0 else ''}", attachments, include
            )
            if lookup:
                ligaLkps.append(lookup)

        ctxLkps, refLkps = self._makeAllContextualMarkLookups(
            tag="mark",
            markAnchors=self.context.contextualMarkToBaseAnchors.items(),
        )

        for context, glyph_anchor_pair in sorted(
            self.context.contextualMarkToLigaAnchors.items(), key=lambda x: -len(x[0])
        ):
            # Group by anchor
            attachments = defaultdict(list)
            for glyphName, anchor in glyph_anchor_pair:
                marks = [[]] * max(
                    a.number
                    for a in self.context.anchorLists[glyphName]
                    if a.key and a.number is not None
                )
                marks[anchor.number - 1] = [anchor]
                attachments[anchor.key].append(MarkToLigaPos(glyphName, marks))
            self._makeContextualMarkLookup(
                attachments,
                context,
                refLkps,
                ctxLkps,
            )

        ctxLkps = list(ctxLkps.values())
        if not baseLkps and not ligaLkps and not ctxLkps:
            return None, []

        feature = ast.FeatureBlock("mark")
        if ctxLkps:
            # When we have contextual lookups, we need to make sure that the
            # contextual and non-contextual lookups are in the right order
            # and we can’t use nested lookups inside the feature block for
            # the referenced lookups, so we put all lookups outside the feature
            # and use lookup references instead.
            # We should probably always do this, as nested lookups are full of
            # gotchas, but this will require updating many test expectations.
            lookups = baseLkps + ligaLkps + refLkps + ctxLkps
            for lookup in baseLkps + ligaLkps + ctxLkps:
                feature.statements.append(ast.LookupReferenceStatement(lookup))
        else:
            lookups = []
            for lookup in baseLkps + ligaLkps:
                feature.statements.append(lookup)
        return feature, lookups

    def _makeContextualMarkLookup(
        self,
        attachments,
        fullcontext,
        refLkps,
        ctxLkps,
        prefix="ContextualMark",
    ):
        for anchorKey, statements in attachments.items():
            # First make the contextual lookup
            if ";" in fullcontext:
                before, after = fullcontext.split(";")
            else:
                before, after = "", fullcontext
            after = after.strip()
            if before not in ctxLkps:
                ctxLkps[before] = ast.LookupBlock(f"{prefix}Dispatch_{len(ctxLkps)}")
                if before:
                    # I know it's not really a comment but this is the easiest way
                    # to get the lookup flag in there without reparsing it.
                    ctxLkps[before].statements.append(ast.Comment(f"{before};"))
            ctxLkp = ctxLkps[before]
            ctxLkp.statements.append(ast.Comment(f"# {after}"))

            # Insert mark glyph names after base glyph names if not specified otherwise.
            if "&" not in after:
                after = after.replace("*", "* &")

            baseGlyphNames = " ".join([s.name for s in statements])
            marks = ast.MarkClassName(self.context.markClasses[anchorKey]).asFea()

            # Replace * with base glyph names
            contextual = after.replace("*", f"[{baseGlyphNames}]")

            # Replace & with mark glyph names
            refLkpName = f"{prefix}_{len(refLkps)}"
            contextual = contextual.replace("&", f"{marks}' lookup {refLkpName}")
            ctxLkp.statements.append(ast.Comment(f"pos {contextual};"))

            # Then make the non-contextual lookup it references
            refLkp = ast.LookupBlock(refLkpName)
            refLkp.statements = [s.asAST() for s in statements]
            refLkps.append(refLkp)

    def _makeMkmkFeature(self, include):
        # First make the non-contextual lookups
        markLkps = []
        for anchorName, attachments in sorted(
            self.context.markToMarkAttachments.items()
        ):
            lkp = self._makeMarkToMarkLookup(anchorName, attachments, include)
            if lkp is not None:
                markLkps.append(lkp)

        ctxLkps, refLkps = self._makeAllContextualMarkLookups(
            tag="mkmk",
            markAnchors=self.context.contextualMarkToMarkAnchors.items(),
            prefix="ContextualMarkToMark",
        )

        ctxLkps = list(ctxLkps.values())
        if not markLkps and not ctxLkps:
            return None, []

        feature = ast.FeatureBlock("mkmk")
        if ctxLkps:
            lookups = markLkps + refLkps + ctxLkps
            for lookup in markLkps + ctxLkps:
                feature.statements.append(ast.LookupReferenceStatement(lookup))
        else:
            lookups = []
            for lookup in markLkps:
                feature.statements.append(lookup)

        return feature, lookups

    def _makeAllContextualMarkLookups(
        self,
        tag,
        markAnchors: list,
        prefix="ContextualMark",
    ):
        markType = None

        if tag in ["mark", "abvm", "blwm"]:
            markType = MarkToBasePos
        elif tag == "mkmk":
            markType = MarkToMarkPos
        else:
            raise AssertionError(tag)

        # Then make the contextual ones
        refLkps = []
        ctxLkps = {}
        # We sort the full context by longest first. This isn't perfect
        # but it gives us the best chance that more specific contexts
        # (typically longer) will take precedence over more general ones.
        for context, glyph_anchor_pair in sorted(markAnchors, key=lambda x: -len(x[0])):
            # Group by anchor
            attachments = defaultdict(list)
            for glyphName, anchor in glyph_anchor_pair:
                attachments[anchor.key].append(markType(glyphName, [anchor]))

            self._makeContextualMarkLookup(
                attachments, context, refLkps, ctxLkps, prefix
            )

        return ctxLkps, refLkps

    def _isAboveMark(self, anchor):
        anchorName = anchor.key if anchor.isContextual else anchor.name

        if anchorName in self.abvmAnchorNames:
            return True
        if (
            anchorName in self.blwmAnchorNames
            or anchorName.startswith("bottom")
            or anchorName.startswith("nukta")
        ):
            return False
        # Glyphs uses (used to use?) a heuristic to guess whether an anchor
        # should go into abvm or blwm. (See
        # https://github.com/googlefonts/ufo2ft/issues/179#issuecomment-390391382)
        # However, this causes issues in variable fonts where an anchor in one
        # master is assigned to a different feature from the same anchor in
        # another master if the Y-coordinates happen to straddle the threshold
        # coordinate. For simplicity, we just place all unknown anchors into
        # the abvm feature.
        return True

    def _isBelowMark(self, anchor):
        return not self._isAboveMark(anchor)

    def _makeAbvmOrBlwmFeature(self, tag, include):
        if tag == "abvm":
            marksFilter = self._isAboveMark
        elif tag == "blwm":
            marksFilter = self._isBelowMark
        else:
            raise AssertionError(tag)

        baseLkps = []
        for attachments in self.context.groupedMarkToBaseAttachments:
            i = len(baseLkps)
            lookup = self._makeMarkLookup(
                f"{tag}_mark2base{'_' + str(i) if i > 0 else ''}",
                attachments,
                include=include,
                marksFilter=marksFilter,
            )
            if lookup:
                baseLkps.append(lookup)
        ligaLkps = []
        for attachments in self.context.groupedMarkToLigaAttachments:
            i = len(ligaLkps)
            lookup = self._makeMarkLookup(
                f"{tag}_mark2liga{'_' + str(i) if i > 0 else ''}",
                attachments,
                include=include,
                marksFilter=marksFilter,
            )
            if lookup:
                ligaLkps.append(lookup)
        mkmkLookups = []
        for anchorName, attachments in sorted(
            self.context.markToMarkAttachments.items()
        ):
            lkp = self._makeMarkToMarkLookup(
                anchorName,
                attachments,
                include=include,
                marksFilter=marksFilter,
                featureTag=tag,
            )
            if lkp is not None:
                mkmkLookups.append(lkp)

        anchorsList = []

        # Let's filter only the corresponding anchors in the current feature (abvm or blwm).
        for (
            context,
            glyph_anchor_pair,
        ) in self.context.contextualMarkAbvmBlwmAnchors.items():
            for _, anchor in glyph_anchor_pair:
                if marksFilter(anchor):
                    anchorsList.append((context, glyph_anchor_pair))

        (ctxLkps, refLkps) = self._makeAllContextualMarkLookups(
            tag=tag,
            markAnchors=anchorsList,
            prefix=f"Contextual{tag.capitalize()}",
        )

        ctxLkps = list(ctxLkps.values())

        if not any([baseLkps, ligaLkps, mkmkLookups]) and not ctxLkps:
            return None, []

        feature = ast.FeatureBlock(tag)
        lookups = []
        if ctxLkps:
            # When we have contextual lookups, we need to make sure that the
            # contextual and non-contextual lookups are in the right order
            # and we can’t use nested lookups inside the feature block for
            # the referenced lookups, so we put all lookups outside the feature
            # and use lookup references instead.
            # We should probably always do this, as nested lookups are full of
            # gotchas, but this will require updating many test expectations.
            lookups = baseLkps + ligaLkps + refLkps + ctxLkps
            for lookup in baseLkps + ligaLkps + ctxLkps:
                feature.statements.append(ast.LookupReferenceStatement(lookup))
        else:
            for lookup in baseLkps + ligaLkps:
                feature.statements.append(lookup)

        feature.statements.extend(mkmkLookups)
        return feature, lookups

    def _makeFeatures(self):
        ctx = self.context

        # First do non-contextual lookups
        ctx.groupedMarkToBaseAttachments = self._groupAttachments(
            self._makeMarkToBaseAttachments()
        )
        ctx.groupedMarkToLigaAttachments = self._groupAttachments(
            self._makeMarkToLigaAttachments()
        )
        ctx.markToMarkAttachments = self._makeMarkToMarkAttachments()

        baseClass = self.context.gdefClasses.base
        ligatureClass = self.context.gdefClasses.ligature
        markClass = self.context.gdefClasses.mark
        (
            ctx.contextualMarkToBaseAnchors,
            ctx.contextualMarkToLigaAnchors,
            ctx.contextualMarkToMarkAnchors,
            ctx.contextualMarkAbvmBlwmAnchors,
        ) = self._makeContextualAttachments(baseClass, ligatureClass, markClass)

        abvmGlyphs, notAbvmGlyphs = self._getAbvmGlyphs()

        def isAbvm(glyphName):
            return glyphName in abvmGlyphs

        def isNotAbvm(glyphName):
            return glyphName in notAbvmGlyphs

        features = {}
        lookups = []
        todo = ctx.todo
        if "mark" in todo:
            mark, markLookups = self._makeMarkFeature(include=isNotAbvm)
            if mark is not None:
                features["mark"] = mark
                lookups.extend(markLookups)
        if "mkmk" in todo:
            mkmk, mkmkLookups = self._makeMkmkFeature(include=isNotAbvm)
            if mkmk is not None:
                features["mkmk"] = mkmk
                lookups.extend(mkmkLookups)
        if "abvm" in todo or "blwm" in todo:
            if abvmGlyphs:
                for tag in ("abvm", "blwm"):
                    if tag not in todo:
                        continue
                    feature, featureLookups = self._makeAbvmOrBlwmFeature(
                        tag, include=isAbvm
                    )
                    if feature is not None:
                        features[tag] = feature
                        lookups.extend(featureLookups)

        return features, lookups

    def _getAbvmGlyphs(self):
        glyphSet = set(self.getOrderedGlyphSet().keys())
        scriptsUsingAbvm = self.scriptsUsingAbvm
        if self.context.feaScripts:
            # https://github.com/googlefonts/ufo2ft/issues/579 Some characters
            # can be used in multiple scripts and some of these scripts might
            # need an abvm feature and some might not, so we filter-out the
            # abvm scripts that the font does not intend to support.
            scriptsUsingAbvm = scriptsUsingAbvm & self.context.feaScripts
        if scriptsUsingAbvm:
            cmap = self.makeUnicodeToGlyphNameMapping()
            unicodeIsAbvm = partial(unicodeInScripts, scripts=scriptsUsingAbvm)

            def unicodeIsNotAbvm(uv):
                return bool(unicodeScriptExtensions(uv) - self.scriptsUsingAbvm)

            if any(unicodeIsAbvm(uv) for uv in cmap):
                # If there are any characters from Indic/USE/Khmer scripts in
                # the cmap, we compile a temporary GSUB table to resolve
                # substitutions and get the set of all the relevant glyphs,
                # including alternate glyphs.
                gsub = self.compileGSUB()
                extras = self.extraSubstitutions()
                glyphGroups = classifyGlyphs(unicodeIsAbvm, cmap, gsub, extras)
                # the 'glyphGroups' dict is keyed by the return value of the
                # classifying include, so here 'True' means all the
                # Indic/USE/Khmer glyphs
                abvmGlyphs = glyphGroups.get(True, set())

                # If a character can be used in Indic/USE/Khmer scripts as well
                # as other scripts, we want to return it in both 'abvmGlyphs'
                # (done above) and 'notAbvmGlyphs' (done below) sets.
                glyphGroups = classifyGlyphs(unicodeIsNotAbvm, cmap, gsub, extras)
                notAbvmGlyphs = glyphGroups.get(True, set())
                # Since cmap might not cover all glyphs, we union with the
                # glyph set.
                notAbvmGlyphs |= glyphSet - abvmGlyphs
                return abvmGlyphs, notAbvmGlyphs
        return set(), glyphSet

    def _write(self):
        self._pruneUnusedAnchors()

        newClassDefs = self._makeMarkClassDefinitions()
        self._setBaseAnchorMarkClasses()

        features, lookups = self._makeFeatures()
        if not features:
            return False

        feaFile = self.context.feaFile

        self._insert(
            feaFile=feaFile,
            markClassDefs=newClassDefs,
            features=[features[tag] for tag in sorted(features.keys())],
            lookups=lookups,
        )

        return True
